r"""
Sub-matrices of linear operators
================================

This tutorial explains how to create linear operators that correspond to a sub-matrix
of another linear operator.

Specifically, given the linear operator :code:`A`, we are
interested in constructing the linear operator that corresponds to its sub-matrix
:code:`A[row_idxs, :][:, col_idxs]`, where :code:`row_idxs` contains the sub-matrix's
row indices, and :code:`col_idxs` contains the sub-matrix's column indices.

First, the imports.
"""

from time import time
from typing import List

import numpy
import torch
from torch import nn

from curvlinops import HessianLinearOperator
from curvlinops.examples.functorch import functorch_hessian
from curvlinops.examples.utils import report_nonclose
from curvlinops.submatrix import SubmatrixLinearOperator

# make deterministic
torch.manual_seed(0)
numpy.random.seed(0)

# %%
#
# Setup
# -----
#
# Let's create some toy data, a small MLP, and use mean-squared error as loss function.

N = 4
D_in = 7
D_hidden = 5
D_out = 3

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")

X1, y1 = torch.rand(N, D_in).to(DEVICE), torch.rand(N, D_out).to(DEVICE)
X2, y2 = torch.rand(N, D_in).to(DEVICE), torch.rand(N, D_out).to(DEVICE)
data = [(X1, y1), (X2, y2)]

model = nn.Sequential(
    nn.Linear(D_in, D_hidden),
    nn.ReLU(),
    nn.Linear(D_hidden, D_hidden),
    nn.Sigmoid(),
    nn.Linear(D_hidden, D_out),
).to(DEVICE)
params = [p for p in model.parameters() if p.requires_grad]

loss_function = nn.MSELoss(reduction="mean").to(DEVICE)

# %%
#
# We will investigate the Hessian. To make sure our results are correct, let's keep
# a Hessian matrix computed via :mod:`functorch` around.

H_functorch = (
    functorch_hessian(model, loss_function, params, data).detach().cpu().numpy()
)

# %%
#
# Here is the corresponding linear operator and a quick check that builds up
# its matrix representation through multiplication with the identity matrix,
# followed by comparison to the Hessian matrix computed via :mod:`functorch`.

H = HessianLinearOperator(model, loss_function, params, data)

report_nonclose(H_functorch, H @ numpy.eye(H.shape[1]))

# %%
#
# Diagonal blocks
# ---------------
#
# The Hessian consists of blocks :code:`(i, j)` that contain the second-order
# derivatives of the loss w.r.t. the parameters in :code:`(params[i], params[j])`.
#
# Let's define a function to extract these blocks from the Hessian:


def extract_block(
    mat: numpy.ndarray, params: List[torch.Tensor], i: int, j: int
) -> numpy.ndarray:
    """Extract the Hessian block from parameters ``i`` and ``j``.

    Args:
        mat: The matrix with block structure.
        params: The parameters defining the blocks.
        i: Row index of the block to be extracted.
        j: Column index of the block to be extracted.

    Returns:
        Block ``(i, j)``. Has shape ``[params[i].numel(), params[j].numel()]``.
    """
    param_dims = [p.numel() for p in params]
    row_start, row_end = sum(param_dims[:i]), sum(param_dims[: i + 1])
    col_start, col_end = sum(param_dims[:j]), sum(param_dims[: j + 1])

    return mat[row_start:row_end, :][:, col_start:col_end]


# %%
#
# As an example, let's extract the block that corresponds to the Hessian w.r.t.
# the first layer's weights in our model.

i, j = 0, 0
H_param0_functorch = extract_block(H_functorch, params, i, j)

# %%
#
# We can build a linear operator for this sub-Hessian by only providing the
# first layer's weight as parameter:

H_param0 = HessianLinearOperator(model, loss_function, [params[i]], data)

# %%
#
# Like this we can get blocks from the diagonal.
#
# Let's check that this linear operator works as expected by multiplying it
# onto the identity matrix and comparing the result to the block we extracted
# from our ground truth:

report_nonclose(H_param0_functorch, H_param0 @ numpy.eye(params[i].numel()))

# %%
#
# Now you might be wondering if we can also build up linear operators for
# off-diagonal blocks. These blocks contain mixed second-order derivatives and
# are not Hessians anymore. For instance, such a block is rectangular in
# general, and thus non-symmetric. Since we are not asking for a Hessian
# anymore, we cannot use the interface of :class:`HessianLinearOperator`.
#
# Luckily, there is a different way to achieve this.
#
# Off-diagonal blocks
# -------------------
#
# As an example,let's try to extract the Hessian block from the first and
# second parameters in our network (i.e. the weights and biases in the first
# layer). For that we need to slice the Hessian differently along its rows and
# columns. We can use the :class:`curvlinops.SubmatrixLinearOperator` class for
# that:

param_dims = [p.numel() for p in params]
i, j = 0, 1
row_start, row_end = sum(param_dims[:i]), sum(param_dims[: i + 1])
col_start, col_end = sum(param_dims[:j]), sum(param_dims[: j + 1])

row_idxs = list(range(row_start, row_end))  # keep the following row indices
col_idxs = list(range(col_start, col_end))  # keep the following column indices

H_param0_param1 = SubmatrixLinearOperator(H, row_idxs, col_idxs)

# %%
#
# As the following test shows, this linear operator indeed represents the
# desired rectangular Hessian block:

H_param0_param1_functorch = extract_block(H_functorch, params, i, j)

report_nonclose(
    H_param0_param1_functorch,
    H_param0_param1_functorch @ numpy.eye(param_dims[j]),
)

# %%
#
# Arbitrary sub-matrices
# ----------------------
#
# So far, we were constrained to blocks spanned by parameter tensors rather
# than arbitrary elements. As the name :class:`SubmatrixLinearOperator`
# suggests, we can use it to create arbitrary sub-matrices.
#
# As an example, let's say we want to keep rows :code:`[0, 13, 42]` of the
# Hessian, and columns :code:`[1, 2, 3]`. This works as follows:

row_idxs = [0, 13, 42]  # keep the following row indices
col_idxs = [1, 2, 3]  # keep the following column indices

H_sub = SubmatrixLinearOperator(H, row_idxs, col_idxs)
H_sub_functorch = H_functorch[row_idxs, :][:, col_idxs]

# %%
#
# Quick check to see if it worked:

report_nonclose(H_sub_functorch, H_sub @ numpy.eye(len(col_idxs)))

# %%
#
# Looks good.
#
# Performance remarks
# ----------------------
#
# By the way, using this interface, we could have also constructed the first
# parameter's Hessian as follows:

i, j = 0, 0
row_start, row_end = sum(param_dims[:i]), sum(param_dims[: i + 1])
col_start, col_end = sum(param_dims[:j]), sum(param_dims[: j + 1])

row_idxs = list(range(row_start, row_end))
col_idxs = list(range(col_start, col_end))

H_param0_alternative = SubmatrixLinearOperator(H, row_idxs, col_idxs)

report_nonclose(H_param0_functorch, H_param0_alternative @ numpy.eye(param_dims[0]))

# %%
#
# In general though, it is a good idea to first reduce the linear operator's
# size as much as possible (in our case, by restricting the parameters to the
# necessary ones using the :code:`params` argument in
# :class:`HessianLinearOperator`) and apply slicing afterwards to save
# computations.
#
# In our example, the matrix-vector product of :code:`H_param0` should
# therefore be faster than that of :code:`H_param0_alternative`:

x = numpy.random.rand(param_dims[0])

# less computations
start = time()
H_param0 @ x
end = time()
print(f"H_param0.matvec: {end - start:.2e} s")

# more computations
start = time()
H_param0_alternative @ x
end = time()
print(f"H_param0_alternative.matvec: {end - start:.2e} s")

# %%
#
# That's all for now.



"""Implements slices of linear operators."""

from __future__ import annotations

from typing import List

from numpy import column_stack, ndarray, zeros
from scipy.sparse.linalg import LinearOperator


class SubmatrixLinearOperator(LinearOperator):
    """Class for sub-matrices of linear operators."""

    def __init__(self, A: LinearOperator, row_idxs: List[int], col_idxs: List[int]):
        """Store the linear operator and indices of its sub-matrix.

        Represents the sub-matrix ``A[row_idxs, :][col_idxs, :]``.

        Args:
            A: A linear operator.
            row_idxs: The sub-matrix's row indices.
            col_idxs: The sub-matrix's column indices.
        """
        self._A = A
        self.set_submatrix(row_idxs, col_idxs)

    def set_submatrix(self, row_idxs: List[int], col_idxs: List[int]):
        """Define the sub-matrix.

        Internally sets the linear operator's shape.

        Args:
            row_idxs: The sub-matrix's row indices.
            col_idxs: The sub-matrix's column indices.

        Raises:
            ValueError: If the index lists contain duplicate values, non-integers,
                or out-of-bounds indices.
        """
        shape = []

        for ax_idx, idxs in enumerate([row_idxs, col_idxs]):
            if any(not isinstance(i, int) for i in idxs):
                raise ValueError("Index lists must contain integers.")
            if len(idxs) != len(set(idxs)):
                raise ValueError("Index lists cannot contain duplicates.")
            if any(i < 0 or i >= self._A.shape[ax_idx] for i in idxs):
                raise ValueError("Index lists contain out-of-bounds indices.")
            shape.append(len(idxs))

        super().__init__(self._A.dtype, shape)
        self._row_idxs = row_idxs
        self._col_idxs = col_idxs

    def _matvec(self, x: ndarray) -> ndarray:
        """Multiply x by the sub-matrix of A.

        Args:
             x: Vector for multiplication. Has shape ``[len(col_idxs)]``.

        Returns:
             Result of the (sub-matrix)-vector-multiplication,
             ``A[row_idxs, :][:, col_idxs] @ x``. Has shape ``[len(row_idxs)]``.
        """
        v = zeros((self._A.shape[1],), dtype=self._A.dtype)
        v[self._col_idxs] = x
        Av = self._A @ v

        return Av[self._row_idxs]

    def _matmat(self, X: ndarray) -> ndarray:
        """Multiply each column of X by the sub-matrix of A.

        Args:
            X: Matrix for multiplication. Has shape ``[len(col_idxs), N]`` with
                abitrary ``N``.

        Returns:
            Result of the (sub-matrix)-matrix-multiplication,
            ``A[row_idxs, :][:, col_idxs] @ x``. Has shape ``[len(row_idxs), N]``.
        """
        return column_stack([self @ col for col in X.T])

    def _adjoint(self) -> SubmatrixLinearOperator:
        """Return the adjoint of the sub-matrix.

        For that, we need to take the adjoint operator, and swap row and column indices.

        Returns:
            The linear operator for the adjoint sub-matrix.
        """
        return type(self)(self._A.adjoint(), self._col_idxs, self._row_idxs)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "CurvLinOps"
copyright = "2022, F. Dangel, L. Tatzel, R. Eschenhagen"
author = "F. Dangel, L. Tatzel, R. Eschenhagen"

# The full version, including alpha/beta/rc tags
release = "1.2.0"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx_gallery.gen_gallery",
    "sphinx.ext.viewcode",  # show source code links
]

# -- Intersphinx config -----------------------------------------------------

intersphinx_mapping = {
    "torch": ("https://pytorch.org/docs/stable/", None),
    "backpack": ("https://docs.backpack.pt/en/master", None),
    "scipy": ("http://docs.scipy.org/doc/scipy/reference/", None),
    "numpy": ("http://docs.scipy.org/doc/numpy/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "curvlinops": ("https://curvlinops.readthedocs.io/en/latest/", None),
}

# -- Sphinx Gallery config ---------------------------------------------------

sphinx_gallery_conf = {
    "examples_dirs": [
        "../examples/basic_usage",
        # "../examples/use_cases",
    ],  # path to your example scripts
    "gallery_dirs": [
        "basic_usage",
        # "use_cases",
    ],  # path to where to save gallery generated output
    "default_thumb_file": "assets/logo.png",
    "filename_pattern": "example",
    "matplotlib_animations": True,
}
# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_logo = "assets/logo.svg"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]



"""Implements linear operators for per-sample Jacobians."""

from __future__ import annotations

from collections.abc import MutableMapping
from typing import Callable, Iterable, List, Optional, Tuple, Union

from backpack.hessianfree.lop import transposed_jacobian_vector_product as vjp
from backpack.hessianfree.rop import jacobian_vector_product as jvp
from numpy import allclose, ndarray
from torch import Tensor, cat, from_numpy, no_grad, stack, zeros_like
from torch.nn import Parameter

from curvlinops._base import _LinearOperator


class JacobianLinearOperator(_LinearOperator):
    """Linear operator for the Jacobian.

    Can be used with SciPy.
    """

    def __init__(
        self,
        model_func: Callable[[Tensor], Tensor],
        params: List[Parameter],
        data: Iterable[Tuple[Union[Tensor, MutableMapping], Tensor]],
        progressbar: bool = False,
        check_deterministic: bool = True,
        num_data: Optional[int] = None,
        batch_size_fn: Optional[Callable[[MutableMapping], int]] = None,
    ):
        r"""Linear operator for the Jacobian as SciPy linear operator.

        Consider a model :math:`f(\mathbf{x}, \mathbf{\theta}): \mathbb{R}^M
        \times \mathbb{R}^D \to \mathbb{R}^C` with parameters
        :math:`\mathbf{\theta}` and input :math:`\mathbf{x}`. Assume we are
        given a data set :math:`\mathcal{D} = \{ (\mathbf{x}_n, \mathbf{y}_n)
        \}_{n=1}^N` of input-target pairs via batches. The model's Jacobian
        :math:`\mathbf{J}_\mathbf{\theta}\mathbf{f}` is an :math:`NC \times D`
        matrix with elements

        .. math::
            \left[
                \mathbf{J}_\mathbf{\theta}\mathbf{f}
            \right]_{(n,c), d}
            =
            \frac{\partial [f(\mathbf{x}_n, \mathbf{\theta})]_c}{\partial \theta_d}\,.

        Note that the data must be supplied in deterministic order.

        Args:
            model_func: Neural network function.
            params: Neural network parameters.
            data: Iterable of batched input-target pairs.
            progressbar: Show progress bar.
            check_deterministic: Check if model and data are deterministic.
            num_data: Number of data points. If ``None``, it is inferred from the data
                at the cost of one traversal through the data loader.
            batch_size_fn: If the ``X``'s in ``data`` are not ``torch.Tensor``, this
                needs to be specified. The intended behavior is to consume the first
                entry of the iterates from ``data`` and return their batch size.
        """
        _batch_size_fn = (
            (lambda X: X.shape[0]) if batch_size_fn is None else batch_size_fn
        )
        num_data = (
            sum(_batch_size_fn(t) for t, _ in data) if num_data is None else num_data
        )
        x = next(iter(data))[0]

        if isinstance(x, Tensor):
            x = x.to(self._infer_device(params))

        num_outputs = model_func(x).shape[1:].numel()
        num_params = sum(p.numel() for p in params)

        super().__init__(
            model_func,
            None,
            params,
            data,
            progressbar=progressbar,
            check_deterministic=check_deterministic,
            shape=(num_data * num_outputs, num_params),
            num_data=num_data,
            batch_size_fn=batch_size_fn,
        )

    def _check_deterministic(self):
        """Verify that the linear operator is deterministic.

        In addition to the checks from the base class, checks that the model
        predictions and data are always the same (loaded in the same order, and
        only deterministic operations in the network.

        Note:
            Deterministic checks should be performed on CPU. We noticed that even when
            it passes on CPU, it can fail on GPU; probably due to non-deterministic
            operations.

        Raises:
            RuntimeError: If the linear operator is not deterministic.
        """
        super()._check_deterministic()

        rtol, atol = 5e-5, 1e-6

        def check_X_y(X1, X2, y1, y2):
            if not allclose(X1, X2) or not allclose(y1, y2):
                self.print_nonclose(X1, X2, rtol=rtol, atol=atol)
                self.print_nonclose(y1, y2, rtol=rtol, atol=atol)
                raise RuntimeError("Non-deterministic data loading detected.")

        with no_grad():
            for (X1, y1), (X2, y2) in zip(
                self._loop_over_data(desc="_check_deterministic_data_pred"),
                self._loop_over_data(desc="_check_deterministic_data_pred2"),
            ):
                pred1, y1 = self._model_func(X1).cpu().numpy(), y1.cpu().numpy()
                pred2, y2 = self._model_func(X2).cpu().numpy(), y2.cpu().numpy()

                if isinstance(X1, Tensor) or isinstance(X2, Tensor):
                    X1, X2 = X1.cpu().numpy(), X2.cpu().numpy()
                    check_X_y(X1, X2, y1, y2)
                else:  # X is a MutableMapping
                    for k in X1.keys():
                        v1, v2 = X1[k], X2[k]

                        if isinstance(v1, Tensor) or isinstance(v2, Tensor):
                            X1, X2 = v1.cpu().numpy(), v2.cpu().numpy()

                        check_X_y(X1, X2, y1, y2)

                if not allclose(pred1, pred2):
                    self.print_nonclose(pred1, pred2, rtol=rtol, atol=atol)
                    raise RuntimeError("Non-deterministic model detected.")

    def _matmat(self, M: ndarray) -> ndarray:
        """Apply the Jacobian to a matrix.

        Args:
            M: Matrix for multiplication. Has shape ``[D, K]``.

        Returns:
            Matrix-multiplication result ``J @ M``. Has shape ``[N * C, K]``.
        """
        num_vecs = M.shape[1]
        M_list = self._preprocess(M)

        result_list = []

        for X, _ in self._loop_over_data(desc="_matmat"):
            output = self._model_func(X)

            # multiply the mini-batch Jacobian onto all vectors
            col = []
            for n in range(num_vecs):
                (col_n,) = jvp(output, self._params, [M[n] for M in M_list])
                col.append(col_n.flatten(start_dim=1))
            # combine columns into a single tensor and append
            result_list.append(stack(col))

        # concatenate over batches
        result_list = [cat(result_list, dim=1)]

        return self._postprocess(result_list)

    def _adjoint(self) -> TransposedJacobianLinearOperator:
        """Return a linear operator representing the adjoint.

        Returns:
            Linear operator representing the transposed Jacobian.
        """
        return TransposedJacobianLinearOperator(
            self._model_func,
            self._params,
            self._data,
            progressbar=self._progressbar,
            check_deterministic=False,
            batch_size_fn=self._batch_size_fn,
        )


class TransposedJacobianLinearOperator(_LinearOperator):
    """Linear operator for the transpose Jacobian.

    Can be used with SciPy.
    """

    def __init__(
        self,
        model_func: Callable[[Tensor], Tensor],
        params: List[Parameter],
        data: Iterable[Tuple[Union[Tensor, MutableMapping], Tensor]],
        progressbar: bool = False,
        check_deterministic: bool = True,
        num_data: Optional[int] = None,
        batch_size_fn: Optional[Callable[[MutableMapping], int]] = None,
    ):
        r"""Linear operator for the transpose Jacobian as SciPy linear operator.

        Consider a model :math:`f(\mathbf{x}, \mathbf{\theta}): \mathbb{R}^M
        \times \mathbb{R}^D \to \mathbb{R}^C` with parameters
        :math:`\mathbf{\theta}` and input :math:`\mathbf{x}`. Assume we are
        given a data set :math:`\mathcal{D} = \{ (\mathbf{x}_n, \mathbf{y}_n)
        \}_{n=1}^N` of input-target pairs via batches. The model's transpose
        Jacobian :math:`(\mathbf{J}_\mathbf{\theta}\mathbf{f})^\top` is an
        :math:`D \times NC` matrix with elements

        .. math::
            \left[
                (\mathbf{J}_\mathbf{\theta}\mathbf{f})^\top
            \right]_{d, (n,c)}
            =
            \frac{\partial [f(\mathbf{x}_n, \mathbf{\theta})]_c}{\partial \theta_d}\,.

        Note that the data must be supplied in deterministic order.

        Args:
            model_func: Neural network function.
            params: Neural network parameters.
            data: Iterable of batched input-target pairs.
            progressbar: Show progress bar.
            check_deterministic: Check if model and data are deterministic.
            num_data: Number of data points. If ``None``, it is inferred from the data
                at the cost of one traversal through the data loader.
            batch_size_fn: If the ``X``'s in ``data`` are not ``torch.Tensor``, this
                needs to be specified. The intended behavior is to consume the first
                entry of the iterates from ``data`` and return their batch size.
        """
        _batch_size_fn = (
            (lambda X: X.shape[0]) if batch_size_fn is None else batch_size_fn
        )
        num_data = (
            sum(_batch_size_fn(t) for t, _ in data) if num_data is None else num_data
        )
        x = next(iter(data))[0]

        if isinstance(x, Tensor):
            x = x.to(self._infer_device(params))

        num_outputs = model_func(x).shape[1:].numel()
        num_params = sum(p.numel() for p in params)

        super().__init__(
            model_func,
            None,
            params,
            data,
            progressbar=progressbar,
            check_deterministic=check_deterministic,
            shape=(num_params, num_data * num_outputs),
            num_data=num_data,
            batch_size_fn=batch_size_fn,
        )

    def _check_deterministic(self):
        """Verify that the linear operator is deterministic.

        In addition to the checks from the base class, checks that the model
        predictions and data are always the same (loaded in the same order, and
        only deterministic operations in the network.

        Note:
            Deterministic checks should be performed on CPU. We noticed that even when
            it passes on CPU, it can fail on GPU; probably due to non-deterministic
            operations.

        Raises:
            RuntimeError: If the linear operator is not deterministic.
        """
        super()._check_deterministic()

        rtol, atol = 5e-5, 1e-6

        with no_grad():
            for (X1, y1), (X2, y2) in zip(
                self._loop_over_data(desc="_check_deterministic_data_pred1"),
                self._loop_over_data(desc="_check_deterministic_data_pred2"),
            ):
                pred1, y1 = self._model_func(X1).cpu().numpy(), y1.cpu().numpy()
                pred2, y2 = self._model_func(X2).cpu().numpy(), y2.cpu().numpy()

                def check_X_y(X1, X2, y1, y2):
                    if not allclose(X1, X2) or not allclose(y1, y2):
                        self.print_nonclose(X1, X2, rtol=rtol, atol=atol)
                        self.print_nonclose(y1, y2, rtol=rtol, atol=atol)
                        raise RuntimeError("Non-deterministic data loading detected.")

                with no_grad():
                    for (X1, y1), (X2, y2) in zip(
                        self._loop_over_data(desc="_check_deterministic_data_pred"),
                        self._loop_over_data(desc="_check_deterministic_data_pred2"),
                    ):
                        pred1, y1 = self._model_func(X1).cpu().numpy(), y1.cpu().numpy()
                        pred2, y2 = self._model_func(X2).cpu().numpy(), y2.cpu().numpy()

                        if isinstance(X1, Tensor) or isinstance(X2, Tensor):
                            X1, X2 = X1.cpu().numpy(), X2.cpu().numpy()
                            check_X_y(X1, X2, y1, y2)
                        else:  # X is a MutableMapping
                            for k in X1.keys():
                                v1, v2 = X1[k], X2[k]

                                if isinstance(v1, Tensor) or isinstance(v2, Tensor):
                                    X1, X2 = v1.cpu().numpy(), v2.cpu().numpy()

                                check_X_y(X1, X2, y1, y2)

                if not allclose(pred1, pred2):
                    self.print_nonclose(pred1, pred2, rtol=rtol, atol=atol)
                    raise RuntimeError("Non-deterministic model detected.")

    def _matmat(self, M: ndarray) -> ndarray:
        """Apply the transpose Jacobian to a matrix.

        Args:
            M: Matrix for multiplication. Has shape ``[N C, K]``.

        Returns:
            Matrix-multiplication result ``J^T @ M``. Has shape ``[D, K]``.
        """
        M_torch = from_numpy(M).to(self._device)
        num_vectors = M_torch.shape[1]

        # allocate result tensors
        out_list = []
        for p in self._params:
            repeat = [num_vectors] + [1] * p.ndim
            out_list.append(zeros_like(p).unsqueeze(0).repeat(*repeat))

        processed = 0
        for X, _ in self._loop_over_data(desc="_matmat"):
            pred = self._model_func(X)
            start, end = processed, processed + pred.numel()

            for n in range(num_vectors):
                v = M_torch[start:end, n].reshape_as(pred)
                JT_v = vjp(pred, self._params, v)

                for p_res, p_vjp in zip(out_list, JT_v):
                    p_res[n].add_(p_vjp)

            processed += pred.numel()

        return self._postprocess(out_list)

    def _adjoint(self) -> JacobianLinearOperator:
        """Return a linear operator representing the adjoint.

        Returns:
            Linear operator representing the Jacobian.
        """
        return JacobianLinearOperator(
            self._model_func,
            self._params,
            self._data,
            progressbar=self._progressbar,
            check_deterministic=False,
            batch_size_fn=self._batch_size_fn,
        )



"""Contains tests for ``curvlinops.activation_hessian``."""

from test.cases import DEVICES, DEVICES_IDS
from test.utils import classification_targets

from numpy import eye as numpy_eye
from pytest import mark, raises
from torch import (
    allclose,
    block_diag,
    device,
    einsum,
    eye,
    from_numpy,
    manual_seed,
    rand,
)
from torch.nn import CrossEntropyLoss, Linear, ReLU, Sequential, Sigmoid

from curvlinops.experimental.activation_hessian import (
    ActivationHessianLinearOperator,
    store_activation,
)


@mark.parametrize("dev", DEVICES, ids=DEVICES_IDS)
def test_store_activation(dev: device):
    """Test context that stores the input/output of a layer.

    Args:
        dev: Device on which to run the tests.
    """
    manual_seed(0)
    layers = [
        Linear(10, 8),
        ReLU(),
        Linear(8, 6),
        Sigmoid(),
        Linear(6, 4),
    ]
    layers = [layer.to(dev) for layer in layers]
    model = Sequential(*layers).to(dev)
    X = rand(5, 10, device=dev)

    # compare with manual forward pass
    activation = ("3", "input", 0)  # input to the sigmoid layer
    activation_storage = []
    with store_activation(model, *activation, activation_storage):
        model(X)
    act = activation_storage.pop()
    assert act.shape == (5, 6)
    assert not activation_storage
    truth = layers[2](layers[1](layers[0](X)))
    assert allclose(act, truth)
    # make sure the hooks were removed when computing ``truth``
    assert not activation_storage

    # check failure scenarios
    # layer name does not exist
    invalid_activation = ("foo", "input", 0)
    with raises(ValueError):
        store_activation(model, *invalid_activation, [])

    # io type does not exist
    invalid_activation = ("3", "foo", 0)
    with raises(ValueError):
        store_activation(model, *invalid_activation, [])

    # destination not empty
    invalid_activation = ("3", "foo", 0)
    with raises(ValueError):
        store_activation(model, *invalid_activation, [42])


@mark.parametrize("dev", DEVICES, ids=DEVICES_IDS)
def test_ActivationHessianLinearOperator(dev: device):
    """Check the Hessian w.r.t. an activation.

    Verifies the Hessian of a function ``l(id(X), y)`` w.r.t. ``X`` where ``id`` is the
    identity.

    Args:
        dev: Device on which to run the tests.
    """
    manual_seed(0)
    batch_size, num_classes = 2, 10

    # model does nothing to the input but needs parameters so the linear
    # operator can infer the device
    model = Linear(num_classes, num_classes, bias=False).to(dev)
    model.weight.data = eye(num_classes)

    loss_func = CrossEntropyLoss(reduction="sum")
    X = rand(batch_size, num_classes, requires_grad=True, device=dev)
    y = classification_targets((batch_size,), num_classes).to(dev)
    data = [(X, y)]
    activation = ("", "input", 0)

    # compute the Hessian matrix representation
    H_linop = ActivationHessianLinearOperator(model, loss_func, activation, data)
    H_mat = from_numpy(H_linop @ numpy_eye(H_linop.shape[1])).to(dev, X.dtype)

    # we know that the Hessian of softmax CE loss is ``diag(p(x)) - p(x) p(x)ᵀ``
    # where ``p(x)`` is the softmax probability on a single datum ``x``. On a batch,
    # the Hessian is the block diagonal stack of these per-sample Hessians
    p = X.softmax(dim=1).detach()
    blocks = []
    for n in range(batch_size):
        p_n = p[n]
        H_n = p_n.diag() - einsum("i,j->ij", p_n, p_n)
        blocks.append(H_n)
    truth = block_diag(*blocks)

    assert allclose(H_mat, truth)



"""Contains tests for ``curvlinops/fisher.py``."""

from collections.abc import MutableMapping
from contextlib import redirect_stdout, suppress

from numpy import random, zeros_like
from pytest import mark, raises

from curvlinops import FisherMCLinearOperator
from curvlinops.examples.functorch import functorch_ggn
from curvlinops.examples.utils import report_nonclose

MAX_REPEATS_MC_SAMPLES = [(10_000, 1), (100, 100)]
MAX_REPEATS_MC_SAMPLES_IDS = [
    f"max_repeats={n}-mc_samples={m}" for (n, m) in MAX_REPEATS_MC_SAMPLES
]
CHECK_EVERY = 100


@mark.montecarlo
@mark.parametrize(
    "max_repeats,mc_samples", MAX_REPEATS_MC_SAMPLES, ids=MAX_REPEATS_MC_SAMPLES_IDS
)
def test_LinearOperator_matvec_expectation(
    case, adjoint: bool, max_repeats: int, mc_samples: int
):
    model_func, loss_func, params, data, batch_size_fn = case

    # Test when X is dict-like but batch_size_fn = None (default)
    if isinstance(data[0][0], MutableMapping):
        with raises(ValueError):
            F = FisherMCLinearOperator(
                model_func,
                loss_func,
                params,
                data,
                mc_samples=mc_samples,
            )

    F = FisherMCLinearOperator(
        model_func,
        loss_func,
        params,
        data,
        batch_size_fn=batch_size_fn,
        mc_samples=mc_samples,
    )
    G_functorch = (
        functorch_ggn(model_func, loss_func, params, data, input_key="x")
        .detach()
        .cpu()
        .numpy()
    )
    if adjoint:
        F, G_functorch = F.adjoint(), G_functorch.conj().T

    x = random.rand(F.shape[1]).astype(F.dtype)
    Gx = G_functorch @ x

    Fx = zeros_like(x)
    atol = 5e-3 * max(abs(Gx))
    rtol = 1e-1

    for m in range(max_repeats):
        Fx += F @ x
        F._seed += 1

        total_samples = (m + 1) * mc_samples
        if total_samples % CHECK_EVERY == 0:
            with redirect_stdout(None), suppress(ValueError):
                report_nonclose(Fx / (m + 1), Gx, rtol=rtol, atol=atol)
                return

    report_nonclose(Fx / max_repeats, Gx, rtol=rtol, atol=atol)


@mark.montecarlo
@mark.parametrize(
    "max_repeats,mc_samples", MAX_REPEATS_MC_SAMPLES, ids=MAX_REPEATS_MC_SAMPLES_IDS
)
def test_LinearOperator_matmat_expectation(
    case, adjoint: bool, max_repeats: int, mc_samples: int, num_vecs: int = 2
):
    model_func, loss_func, params, data, batch_size_fn = case

    F = FisherMCLinearOperator(
        model_func,
        loss_func,
        params,
        data,
        batch_size_fn=batch_size_fn,
        mc_samples=mc_samples,
    )
    G_functorch = (
        functorch_ggn(model_func, loss_func, params, data, input_key="x")
        .detach()
        .cpu()
        .numpy()
    )
    if adjoint:
        F, G_functorch = F.adjoint(), G_functorch.conj().T

    X = random.rand(F.shape[1], num_vecs).astype(F.dtype)
    GX = G_functorch @ X

    FX = zeros_like(X)
    atol = 5e-3 * max(abs(GX.flatten()))
    rtol = 1.5e-1

    for m in range(max_repeats):
        FX += F @ X
        F._seed += 1

        total_samples = (m + 1) * mc_samples
        if total_samples % CHECK_EVERY == 0:
            with redirect_stdout(None), suppress(ValueError):
                report_nonclose(FX / (m + 1), GX, rtol=rtol, atol=atol)
                return

    report_nonclose(FX / max_repeats, GX, rtol=rtol, atol=atol)



"""Contains LinearOperator implementation of the (approximate) Fisher."""

from __future__ import annotations

from collections.abc import MutableMapping
from math import sqrt
from typing import Callable, Iterable, List, Optional, Tuple, Union

from backpack.hessianfree.ggnvp import ggn_vector_product_from_plist
from einops import einsum, rearrange
from numpy import ndarray
from torch import Generator, Tensor, as_tensor, normal, softmax, zeros, zeros_like
from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, MSELoss, Parameter
from torch.nn.functional import one_hot

from curvlinops._base import _LinearOperator


class FisherMCLinearOperator(_LinearOperator):
    r"""Monte-Carlo approximation of the Fisher as SciPy linear operator.

    Consider the empirical risk

    .. math::
        \mathcal{L}(\mathbf{\theta})
        =
        c \sum_{n=1}^{N}
        \ell(f_{\mathbf{\theta}}(\mathbf{x}_n), \mathbf{y}_n)

    with :math:`c = \frac{1}{N}` for ``reduction='mean'`` and :math:`c=1` for
    ``reduction='sum'``. Let :math:`\ell(\mathbf{f}, \mathbf{y}) = - \log
    q(\mathbf{y} \mid \mathbf{f})` be a negative log-likelihood loss. Denoting
    :math:`\mathbf{f}_n = f_{\mathbf{\theta}}(\mathbf{x}_n)`, the Fisher
    information matrix is

    .. math::
        c \sum_{n=1}^{N}
        \left(
            \mathbf{J}_{\mathbf{\theta}}
            \mathbf{f}_n
        \right)^\top
        \mathbb{E}_{\mathbf{\tilde{y}}_n \sim q( \cdot  \mid \mathbf{f}_n)}
        \left[
            \nabla_{\mathbf{f}_n}^2
            \ell(\mathbf{f}_n, \mathbf{\tilde{y}}_n)
        \right]
        \left(
            \mathbf{J}_{\mathbf{\theta}}
            \mathbf{f}_n
        \right)
        \\
        =
        c \sum_{n=1}^{N}
        \left(
            \mathbf{J}_{\mathbf{\theta}}
            \mathbf{f}_n
        \right)^\top
        \mathbb{E}_{\mathbf{\tilde{y}}_n \sim q( \cdot  \mid \mathbf{f}_n)}
        \left[
            \left(
            \nabla_{\mathbf{f}_n}
            \ell(\mathbf{f}_n, \mathbf{\tilde{y}}_n)
            \right)
            \left(
            \nabla_{\mathbf{f}_n}
            \ell(\mathbf{f}_n, \mathbf{\tilde{y}}_n)
            \right)^{\top}
        \right]
        \left(
            \mathbf{J}_{\mathbf{\theta}}
            \mathbf{f}_n
        \right)
        \\
        \approx
        c \sum_{n=1}^{N}
        \left(
            \mathbf{J}_{\mathbf{\theta}}
            \mathbf{f}_n
        \right)^\top
        \frac{1}{M}
        \sum_{m=1}^M
        \left[
            \left(
            \nabla_{\mathbf{f}_n}
            \ell(\mathbf{f}_n, \mathbf{\tilde{y}}_{n}^{(m)})
            \right)
            \left(
            \nabla_{\mathbf{f}_n}
            \ell(\mathbf{f}_n, \mathbf{\tilde{y}}_{n}^{(m)})
            \right)^{\top}
        \right]
        \left(
            \mathbf{J}_{\mathbf{\theta}}
            \mathbf{f}_n
        \right)

    with sampled targets :math:`\mathbf{\tilde{y}}_{n}^{(m)} \sim q( \cdot \mid
    \mathbf{f}_n)`. The expectation over the model's likelihood is approximated
    via a Monte-Carlo estimator with :math:`M` samples.

    The linear operator represents a deterministic sample from this MC Fisher estimator.
    To generate different samples, you have to create instances with varying random
    seed argument.
    """

    supported_losses = (MSELoss, CrossEntropyLoss, BCEWithLogitsLoss)

    def __init__(
        self,
        model_func: Callable[[Tensor], Tensor],
        loss_func: Union[MSELoss, CrossEntropyLoss],
        params: List[Parameter],
        data: Iterable[Tuple[Union[Tensor, MutableMapping], Tensor]],
        progressbar: bool = False,
        check_deterministic: bool = True,
        seed: int = 2147483647,
        mc_samples: int = 1,
        num_data: Optional[int] = None,
        batch_size_fn: Optional[Callable[[MutableMapping], int]] = None,
    ):
        """Linear operator for the MC approximation of the Fisher.

        Note:
            f(X; θ) denotes a neural network, parameterized by θ, that maps a mini-batch
            input X to predictions p. ℓ(p, y) maps the prediction to a loss, using the
            mini-batch labels y.

        Args:
            model_func: A function that maps the mini-batch input X to predictions.
                Could be a PyTorch module representing a neural network.
            loss_func: Loss function criterion. Maps predictions and mini-batch labels
                to a scalar value.
            params: List of differentiable parameters used by the prediction function.
            data: Source from which mini-batches can be drawn, for instance a list of
                mini-batches ``[(X, y), ...]`` or a torch ``DataLoader``. Note that ``X``
                could be a ``dict`` or ``UserDict``; this is useful for custom models.
                In this case, you must (i) specify the ``batch_size_fn`` argument, and
                (ii) take care of preprocessing like ``X.to(device)`` inside of your
                ``model.forward()`` function. Due to the sequential internal Monte-Carlo
                sampling, batches must be presented in the same deterministic
                order (no shuffling!).
            progressbar: Show a progressbar during matrix-multiplication.
                Default: ``False``.
            check_deterministic: Probe that model and data are deterministic, i.e.
                that the data does not use `drop_last` or data augmentation. Also, the
                model's forward pass could depend on the order in which mini-batches
                are presented (BatchNorm, Dropout). Default: ``True``. This is a
                safeguard, only turn it off if you know what you are doing.
            seed: Seed used to construct the internal random number generator used to
                draw samples at the beginning of each matrix-vector product.
                Default: ``2147483647``
            mc_samples: Number of samples to use. Default: ``1``.
            num_data: Number of data points. If ``None``, it is inferred from the data
                at the cost of one traversal through the data loader.
            batch_size_fn: If the ``X``'s in ``data`` are not ``torch.Tensor``, this
                needs to be specified. The intended behavior is to consume the first
                entry of the iterates from ``data`` and return their batch size.

        Raises:
            NotImplementedError: If the loss function differs from ``MSELoss``,
                BCEWithLogitsLoss, or ``CrossEntropyLoss``.
        """
        if not isinstance(loss_func, self.supported_losses):
            raise NotImplementedError(
                f"Loss must be one of {self.supported_losses}. Got: {loss_func}."
            )
        self._seed = seed
        self._generator: Union[None, Generator] = None
        self._mc_samples = mc_samples
        super().__init__(
            model_func,
            loss_func,
            params,
            data,
            progressbar=progressbar,
            check_deterministic=check_deterministic,
            num_data=num_data,
            batch_size_fn=batch_size_fn,
        )

    def _matmat(self, M: ndarray) -> ndarray:
        """Multiply the MC-Fisher onto a matrix.

        Create and seed the random number generator.

        Args:
            M: Matrix for multiplication.

        Returns:
            Matrix-multiplication result ``mat @ M``.
        """
        if self._generator is None or self._generator.device != self._device:
            self._generator = Generator(device=self._device)
        self._generator.manual_seed(self._seed)

        return super()._matmat(M)

    def _matmat_batch(
        self, X: Union[Tensor, MutableMapping], y: Tensor, M_list: List[Tensor]
    ) -> Tuple[Tensor, ...]:
        """Apply the mini-batch MC-Fisher to a matrix.

        Args:
            X: Input to the DNN.
            y: Ground truth.
            M_list: Matrix to be multiplied with in list format.
                Tensors have same shape as trainable model parameters, and an
                additional leading axis for the matrix columns.

        Returns:
            Result of MC-Fisher multiplication in list format. Has the same shape as
            ``M_list``, i.e. each tensor in the list has the shape of a parameter and a
            leading dimension of matrix columns.
        """
        # compute ∂ℓₙ(yₙₘ)/∂fₙ where fₙ is the prediction for datum n and
        # yₙₘ is the m-th sampled label for datum n
        output = self._model_func(X)
        # If >2d output we convert to an equivalent 2d output
        if isinstance(self._loss_func, CrossEntropyLoss):
            output = rearrange(output, "batch c ... -> (batch ...) c")
            y = rearrange(y, "batch ... -> (batch ...)")
        else:
            output = rearrange(output, "batch ... c -> (batch ...) c")
            y = rearrange(y, "batch ... c -> (batch ...) c")

        grad_output = self.sample_grad_output(output, self._mc_samples, y)

        # Adjust the scale depending on the loss function and reduction used
        num_loss_terms, C = output.shape
        reduction_factor = {
            "mean": (
                num_loss_terms
                if isinstance(self._loss_func, CrossEntropyLoss)
                else num_loss_terms * C
            ),
            "sum": 1.0,
        }[self._loss_func.reduction]

        # Compute the pseudo-loss L' := 0.5 / (M * c) ∑ₙ ∑ₘ fₙᵀ (gₙₘ gₙₘᵀ) fₙ where
        # gₙₘ = ∂ℓₙ(yₙₘ)/∂fₙ (detached) and M is the number of MC samples.
        # The GGN of L' linearized at fₙ is the MC Fisher.
        # We can thus multiply with it by computing the GGN-vector products of L'.
        loss = (
            0.5
            / reduction_factor
            / self._mc_samples
            * (einsum(output, grad_output, "n ..., m n ... -> m n") ** 2).sum()
        )

        # Multiply the MC Fisher onto each vector in the input matrix
        result_list = [zeros_like(M) for M in M_list]
        num_vectors = M_list[0].shape[0]
        for v in range(num_vectors):
            for idx, ggnvp in enumerate(
                ggn_vector_product_from_plist(
                    loss, output, self._params, [M[v] for M in M_list]
                )
            ):
                result_list[idx][v].add_(ggnvp.detach())

        return tuple(result_list)

    def sample_grad_output(self, output: Tensor, num_samples: int, y: Tensor) -> Tensor:
        """Draw would-be gradients ``∇_f log p(·|f)``.

        For a single data point, the would-be gradient's outer product equals the
        Hessian ``∇²_f log p(·|f)`` in expectation.

        Currently only supports ``MSELoss``, ``CrossEntropyLoss``, and
        ``BCEWithLogitsLoss``.

        The returned gradient does not account for the scaling of the loss function by
        the output dimension ``C`` that ``MSELoss`` and ``BCEWithLogitsLoss`` apply when
        ``reduction='mean'``.

        Args:
            output: model prediction ``f`` for multiple data with batch axis as
                0th dimension.
            num_samples: Number of samples to draw.
            y: Labels of the data on which output was produced.

        Returns:
            Samples of the gradient w.r.t. the model prediction.
            Has shape ``[num_samples, *output.shape]``.

        Raises:
            NotImplementedError: For unsupported loss functions.
            NotImplementedError: If the prediction does not have two dimensions.
            NotImplementedError: If binary classification labels are not binary.
        """
        if output.ndim != 2:
            raise NotImplementedError(f"Only 2d outputs supported. Got {output.shape}")

        C = output.shape[1]

        if isinstance(self._loss_func, MSELoss):
            std = as_tensor(sqrt(0.5), device=output.device)
            mean = zeros(
                num_samples, *output.shape, device=output.device, dtype=output.dtype
            )
            return 2 * normal(mean, std, generator=self._generator)

        elif isinstance(self._loss_func, CrossEntropyLoss):
            prob = softmax(output, dim=1)
            sample = prob.multinomial(
                num_samples=num_samples, replacement=True, generator=self._generator
            )
            sample = rearrange(sample, "batch s -> s batch")
            onehot_sample = one_hot(sample, num_classes=C)
            # repeat ``num_sample`` times along a new leading axis to avoid broadcasting
            prob = prob.unsqueeze(0).expand_as(onehot_sample)
            return prob - onehot_sample

        elif isinstance(self._loss_func, BCEWithLogitsLoss):
            unique = set(y.unique().flatten().tolist())
            if not unique.issubset({0, 1}):
                raise NotImplementedError(
                    "Only binary targets (0, 1) are currently supported with"
                    + f" BCEWithLogitsLoss. Got {unique}."
                )
            prob = output.sigmoid()
            # repeat ``num_sample`` times along a new leading axis
            prob = prob.unsqueeze(0).expand(num_samples, -1, -1)
            sample = prob.bernoulli(generator=self._generator)
            return prob - sample

        else:
            raise NotImplementedError(f"Supported losses: {self.supported_losses}")

    def _adjoint(self) -> FisherMCLinearOperator:
        """Return the linear operator representing the adjoint.

        The Fisher MC-approximation is real symmetric, and hence self-adjoint.

        Returns:
            Self.
        """
        return self



r"""
Monte-Carlo approximation of the Fisher
=======================================

In this tutorial, we will compare two approaches to compute the Fisher
information matrix:

1. The Fisher as expected Hessian under the model's likelihood coincides with
   the generalized Gauss-Newton (GGN) matrix for common loss functions, like
   :class:`torch.nn.MSELoss` and :class:`torch.nn.CrossEntropyLoss`. For these settings,
   Fisher = GGN.

2. The Fisher can also be seen as expectation of the gradient outer product w.r.t. the
   model's likelihood. This expectation can be approximated by computing
   the outer product of 'would-be' gradients where the loss is evaluated on a
   label sampled from the model's likelihood, rather than the true label.

The first approach is implemented by :class:`curvlinops.GGNLinearOperator`, the
second by :class:`curvlinops.FisherMCLinearOperator`. We will see that both approaches
coincide as the Monte-Carlo approximation converges.

Let's get the imports out of our way.
"""

import matplotlib.pyplot as plt
import numpy
import torch
from matplotlib import animation
from torch import nn

from curvlinops import FisherMCLinearOperator, GGNLinearOperator

# make deterministic
torch.manual_seed(0)
numpy.random.seed(0)

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")

# %%
# Setup
# -----
#
# We will create a synthetic classification task, a small CNN, and use
# cross-entropy error as loss function.

Ns = [4, 6]
D_in = 7
D_hidden = 5
C = 3

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")

data = [
    (
        torch.rand(N, D_in).to(DEVICE),  # X
        torch.randint(low=0, high=C, size=(N,)).to(DEVICE),  # y
    )
    for N in Ns
]

model = nn.Sequential(
    nn.Linear(D_in, D_hidden),
    nn.ReLU(),
    nn.Linear(D_hidden, D_hidden),
    nn.Sigmoid(),
    nn.Linear(D_hidden, C),
).to(DEVICE)
params = [p for p in model.parameters() if p.requires_grad]

loss_function = nn.CrossEntropyLoss(reduction="mean").to(DEVICE)


# %%
# A first comparison
# ------------------
#
# Let's create linear operators for the GGN and the Monte-Carlo approximated
# Fisher and compute their matrix representations by multiplying them onto the
# identity matrix:

GGN = GGNLinearOperator(model, loss_function, params, data)
F = FisherMCLinearOperator(model, loss_function, params, data)

D = GGN.shape[0]
identity = numpy.eye(D)

GGN_mat = GGN @ identity
F_mat = F @ identity

# %%
#
# We can use the residual's Frobenius norm to quantify the approximation error
# of the Monte-Carlo estimator:

residual_norm = numpy.linalg.norm(GGN_mat - F_mat)
print(f"Residual (Frobenius) norm: {residual_norm:.5f}")

# %%
# Setting the number of MC samples
# --------------------------------
#
# To get more accurate estimates, we can use more samples in the MC approximation.
# This is achieved by specifying the optional :code:`mc_samples` argument to
# :class:`curvlinops.FisherMCLinearOperator`. The default value is :code:`1`.
#
# Here are the residual Frobenius norms when using more samples:

mc_samples = [1, 2, 4, 8]
residual_norms = []

for mc in mc_samples:
    F = FisherMCLinearOperator(model, loss_function, params, data, mc_samples=mc)
    F_mat = F @ identity
    residual_norms.append(numpy.linalg.norm(GGN_mat - F_mat))

for mc, norm in zip(mc_samples, residual_norms):
    print(f"mc_samples = {mc},\tresidual (Frobenius) norm = {norm:.5f}")

# %%
# Setting the random seed
# -----------------------
#
# You may have noticed above that the two linear operators created with
# :code:`mc_samples=1` yield identical residual Frobenius norms. This is
# because the two linear operators realize the same matrix, i.e. the same
# sample from the Monte-Carlo estimator.
#
# To see that :class:`curvlinops.FisherMCLinearOperator` indeed represents a
# deterministic matrix, let's create two linear operators with identical
# hyperparameters and compare their matrix representations. After creating the
# first linear operator, we generate some random numbers to show that the
# global random number generator does not influence the Monte-Carlo estimator:

F1_mat = FisherMCLinearOperator(model, loss_function, params, data) @ identity

# draw some random numbers to modify the global random number generator's state
torch.rand(123)

F2_mat = FisherMCLinearOperator(model, loss_function, params, data) @ identity

# still, we get the same deterministic approximation
residual_norm = numpy.linalg.norm(F1_mat - F2_mat)
if numpy.isclose(residual_norm, 0.0):
    print(residual_norm)
else:
    raise RuntimeError(f"Residual Frobenius norm should be 0. Got {residual_norm}.")

# %%
#
# This is because the class uses an internal random number generator to draw
# samples. Therefore, it will not be affected by changes to the global random
# number generator's state.
#
# You can get different realizations of the Monte-Carlo estimator by specifying
# the optional :code:`seed` argument. The above comparison with differently
# seeded linear operators leads to different matrices:

seed1 = 123456
F1_mat = (
    FisherMCLinearOperator(model, loss_function, params, data, seed=seed1) @ identity
)

seed2 = 654321
F2_mat = (
    FisherMCLinearOperator(model, loss_function, params, data, seed=seed2) @ identity
)

# now, we get two different deterministic approximations
residual_norm = numpy.linalg.norm(F1_mat - F2_mat)
if not numpy.isclose(residual_norm, 0.0):
    print(residual_norm)
else:
    raise RuntimeError(f"Residual Frobenius norm should be ≠0. Got {residual_norm}.")

# %%
#
# Approximation quality
# ---------------------
#
# Finally, let's combine what we have seen so far to visualize how well the
# Monte-Carlo approximated Fisher approximates the GGN.
#
# To do that, we will repeatedly draw samples for the Fisher information matrix
# and combine them to yield an estimate that incorporates all previous
# iterations. This approach allows to record snapshots of the estimator at a
# different number of total incorporated MC samples.
#
# We will use a :code:`logspace` for taking snapshots.

num_steps = 25
mc_samples = numpy.unique(
    numpy.logspace(0, 2, num_steps, endpoint=True, dtype=numpy.int32)
)
F_snapshots = []
F_accumulated = numpy.zeros((D, D))
start_seed = 123456789

for seed, mc in enumerate(range(mc_samples.max()), start=start_seed):
    # NOTE Only use `check_deterministic=False` if you know what you are doing
    # We do this here because we have previously convinced ourselves that the created
    # linear operators indeed realize deterministic matrices.
    F = FisherMCLinearOperator(
        model, loss_function, params, data, seed=seed, check_deterministic=False
    )
    F_accumulated += F @ identity
    if mc + 1 in mc_samples:
        F_snapshots.append(F_accumulated / (mc + 1))


# %%
#
# Let's visualize both the residual matrices and their Frobenius norms. To
# visualize the matrix, we will use the element-wise logarithm of its absolute
# value (shifted by a small constant to avoid taking the logarithm of 0):

residual_snapshots = [mat - GGN_mat for mat in F_snapshots]
residual_norms = [numpy.linalg.norm(res) for res in residual_snapshots]


def transform(mat: numpy.ndarray, epsilon: float = 1e-5) -> numpy.ndarray:
    """Transformation applied to the matrix before plotting.

    Applies element-wise absolute value, shifts by epsilon, then takes the
    element-wise logarithm.

    Args:
        mat: Matrix.
        epsilon: Small shift to avoid taking the log of 0.

    Returns:
        Transformed matrix
    """
    return numpy.log10(numpy.abs(mat) + epsilon)


# %%
#
# Here's the plotting code (feel free to skip to the visualization).

img_width = 4
rows, columns = 1, 2
fig, axes = plt.subplots(
    nrows=rows, ncols=columns, figsize=(columns * img_width, rows * img_width)
)
ax_img, ax_fro = axes[0], axes[1]

min_img = min(transform(res).min() for res in residual_snapshots)
max_img = max(transform(res).max() for res in residual_snapshots)

min_fro = 0.0
max_fro = max(residual_norms)

ax_fro.set_xlim(mc_samples.min(), mc_samples.max())
ax_fro.semilogx()
ax_fro.set_xlabel("MC samples")
ax_fro.set_ylabel("residual Frobenius norm")
ax_fro.set_ylim(min_fro, 1.05 * max_fro)
ln_style = "bo"

# collects artists to draw in each frame of the animation
artists = []

for frame_idx in range(len(mc_samples)):
    snapshot = residual_snapshots[frame_idx]
    img = ax_img.imshow(transform(snapshot), vmin=min_img, vmax=max_img, animated=True)

    # workaround for animated title: https://stackoverflow.com/a/47421938
    ax_img_title = plt.text(
        0.5,
        1.01,
        f"Residual ({mc_samples[frame_idx]} samples)",
        horizontalalignment="center",
        verticalalignment="bottom",
        transform=ax_img.transAxes,
    )

    if frame_idx == 0:
        ax_img.imshow(transform(snapshot), vmin=min_img, vmax=max_img)
        plt.colorbar(img, ax=ax_img, label="logarithmic absolute entries", shrink=0.8)
        ax_fro.plot(
            mc_samples[: frame_idx + 1], residual_norms[: frame_idx + 1], ln_style
        )
        plt.subplots_adjust(wspace=0.5, bottom=0.2)

    ln_fro = ax_fro.plot(
        mc_samples[: frame_idx + 1], residual_norms[: frame_idx + 1], ln_style
    )

    artists.append([ax_img_title, img] + ln_fro)


ani = animation.ArtistAnimation(
    fig, artists, interval=1000, blit=False, repeat_delay=1000
)

# %%
#
# Here's a more qualitative comparison that contrasts the GGN matrix and the MC
# approximated Fisher (both transformed to logspace as described above):

img_width = 4
rows, columns = 1, 2
fig, axes = plt.subplots(
    nrows=rows, ncols=columns, figsize=(columns * img_width, rows * img_width)
)

min_value = min(transform(mat).min() for mat in F_snapshots + [GGN_mat])
max_value = max(transform(mat).max() for mat in F_snapshots + [GGN_mat])

# collects artists to draw in each frame of the animation
artists = []

ax_GGN, ax_F = axes[0], axes[1]
ax_GGN.set_title("GGN")


for frame_idx in range(len(mc_samples)):
    im_GGN = ax_GGN.imshow(
        transform(GGN_mat), vmin=min_value, vmax=max_value, animated=True
    )
    im_F = ax_F.imshow(
        transform(F_snapshots[frame_idx]), vmin=min_value, vmax=max_value, animated=True
    )

    # workaround for animated title: https://stackoverflow.com/a/47421938
    ax_F_title = plt.text(
        0.5,
        1.01,
        f"Fisher-MC ({mc_samples[frame_idx]} samples)",
        horizontalalignment="center",
        verticalalignment="bottom",
        transform=ax_F.transAxes,
    )

    if frame_idx == 0:
        img = ax_GGN.imshow(transform(GGN_mat), vmin=min_value, vmax=max_value)
        ax_F.imshow(transform(F_snapshots[frame_idx]), vmin=min_value, vmax=max_value)
        fig.colorbar(
            img,
            ax=axes.ravel().tolist(),
            label="logarithmic absolute entries",
            shrink=0.8,
        )

    artists.append([ax_F_title, im_GGN, im_F])


ani = animation.ArtistAnimation(
    fig, artists, interval=1000, blit=False, repeat_delay=1000
)

# %%
#
# That's all for now.



"""Contains LinearOperator implementation of the Hessian."""

from __future__ import annotations

from collections.abc import MutableMapping
from typing import List, Tuple, Union

from backpack.hessianfree.hvp import hessian_vector_product
from torch import Tensor, zeros_like
from torch.autograd import grad

from curvlinops._base import _LinearOperator
from curvlinops.utils import split_list


class HessianLinearOperator(_LinearOperator):
    r"""Hessian as SciPy linear operator.

    Consider the empirical risk

    .. math::
        \mathcal{L}(\mathbf{\theta})
        =
        c \sum_{n=1}^{N}
        \ell(f_{\mathbf{\theta}}(\mathbf{x}_n), \mathbf{y}_n)

    with :math:`c = \frac{1}{N}` for ``reduction='mean'`` and :math:`c=1` for
    ``reduction='sum'``. The Hessian matrix is

    .. math::
        \nabla^2_{\mathbf{\theta}} \mathcal{L}
        =
        c \sum_{n=1}^{N}
        \nabla_{\mathbf{\theta}}^2
        \ell(f_{\mathbf{\theta}}(\mathbf{x}_n), \mathbf{y}_n)\,.

    Attributes:
        SUPPORTS_BLOCKS: Whether the linear operator supports block operations.
            Default is ``True``.
    """

    SUPPORTS_BLOCKS: bool = True

    def _matmat_batch(
        self, X: Union[Tensor, MutableMapping], y: Tensor, M_list: List[Tensor]
    ) -> Tuple[Tensor, ...]:
        """Apply the mini-batch Hessian to a matrix.

        Args:
            X: Input to the DNN.
            y: Ground truth.
            M_list: Matrix to be multiplied with in list format.
                Tensors have same shape as trainable model parameters, and an
                additional leading axis for the matrix columns.

        Returns:
            Result of Hessian multiplication in list format. Has the same shape as
            ``M_list``, i.e. each tensor in the list has the shape of a parameter and a
            leading dimension of matrix columns.
        """
        loss = self._loss_func(self._model_func(X), y)

        # Re-cycle first backward pass from the HVP's double-backward
        grad_params = grad(loss, self._params, create_graph=True)

        num_vecs = M_list[0].shape[0]
        result = [zeros_like(M) for M in M_list]

        # per-block HMP
        for M_block, p_block, g_block, res_block in zip(
            split_list(M_list, self._block_sizes),
            split_list(self._params, self._block_sizes),
            split_list(grad_params, self._block_sizes),
            split_list(result, self._block_sizes),
        ):
            for n in range(num_vecs):
                col_n = hessian_vector_product(
                    loss, p_block, [M[n] for M in M_block], grad_params=g_block
                )
                for p, col in enumerate(col_n):
                    res_block[p][n].add_(col)

        return tuple(result)

    def _adjoint(self) -> HessianLinearOperator:
        """Return the linear operator representing the adjoint.

        The Hessian is real symmetric, and hence self-adjoint.

        Returns:
            Self.
        """
        return self



"""Contains functorch functionality for the examples."""

from collections.abc import MutableMapping
from math import sqrt
from typing import Dict, Iterable, List, Optional, Tuple, Union

import torch
from einops import rearrange
from torch import Tensor, cat, stack
from torch.func import functional_call, grad, hessian, jacrev, jvp
from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, Module, MSELoss


def blocks_to_matrix(blocks: Dict[str, Dict[str, Tensor]]) -> Tensor:
    """Convert a block representation into a matrix.

    Assumes the diagonal blocks to be quadratic to automatically detect their dimension.

    Args:
        blocks: Nested dictionaries with keys denoting parameter names, and blocks
            ``(i, j)`` denoting the matrix block w.r.t parameters ``[i, j]``.

    Returns:
        Two-dimensional matrix with concatenated and flattened blocks.
    """
    row_blocks = []

    param_names = list(blocks.keys())
    for name, param_block in blocks.items():
        num_rows = int(sqrt(param_block[name].numel()))
        col_blocks = [param_block[n].reshape(num_rows, -1) for n in param_names]
        row_blocks.append(cat(col_blocks, dim=1))

    return cat(row_blocks)


def functorch_hessian(
    model_func: Module,
    loss_func: Module,
    params: List[Tensor],
    data: Iterable[Tuple[Union[Tensor, MutableMapping], Tensor]],
    input_key: Optional[str] = None,
) -> Tensor:
    """Compute the Hessian with functorch.

    Args:
        model_func: A function that maps the mini-batch input X to predictions.
            Could be a PyTorch module representing a neural network.
        loss_func: Loss function criterion. Maps predictions and mini-batch labels
            to a scalar value.
        params: List of differentiable parameters used by the prediction function.
        data: Source from which mini-batches can be drawn, for instance a list of
            mini-batches ``[(X, y), ...]`` or a torch ``DataLoader``.
        input_key: Key to obtain the input tensor when ``X`` is a dict-like object.

    Returns:
        Square matrix containing the Hessian.
    """
    (dev,) = {p.device for p in params}
    X, y = _concatenate_batches(data, input_key, device=dev)
    params_dict = _make_params_dict(model_func, params)

    def loss(
        X: Union[Tensor, MutableMapping], y: Tensor, params_dict: Dict[str, Tensor]
    ) -> Tensor:
        """Compute the loss given a mini-batch and the neural network parameters.

        # noqa: DAR101
        # noqa: DAR201
        """
        output = functional_call(model_func, params_dict, X)
        return functional_call(loss_func, {}, (output, y))

    params_argnum = 2
    hessian_fn = hessian(loss, argnums=params_argnum)

    return blocks_to_matrix(hessian_fn(X, y, params_dict))


def functorch_ggn(
    model_func: Module,
    loss_func: Module,
    params: List[Tensor],
    data: Iterable[Tuple[Union[Tensor, MutableMapping], Tensor]],
    input_key: Optional[str] = None,
) -> Tensor:
    """Compute the GGN with functorch.

    The GGN is the Hessian when the model is replaced by its linearization.

    Args:
        model_func: A function that maps the mini-batch input X to predictions.
            Could be a PyTorch module representing a neural network.
        loss_func: Loss function criterion. Maps predictions and mini-batch labels
            to a scalar value.
        params: List of differentiable parameters used by the prediction function.
        data: Source from which mini-batches can be drawn, for instance a list of
            mini-batches ``[(X, y), ...]`` or a torch ``DataLoader``.
        input_key: Key to obtain the input tensor when ``X`` is a dict-like object.

    Returns:
        Square matrix containing the GGN.
    """
    (dev,) = {p.device for p in params}
    X, y = _concatenate_batches(data, input_key, device=dev)
    params_dict = _make_params_dict(model_func, params)

    def linearized_model(
        anchor_dict: Dict[str, Tensor],
        params_dict: Dict[str, Tensor],
        X: Union[Tensor, MutableMapping],
    ) -> Tensor:
        """Evaluate the model at params, using its linearization around anchor.

        # noqa: DAR101
        # noqa: DAR201
        """

        def model_fn_params_only(params_dict: Dict[str, Tensor]) -> Tensor:
            return functional_call(model_func, params_dict, X)

        diff_dict = {n: params_dict[n] - anchor_dict[n] for n in params_dict}
        model_at_anchor, jvp_diff = jvp(
            model_fn_params_only, (anchor_dict,), (diff_dict,)
        )

        return model_at_anchor + jvp_diff

    def linearized_loss(
        X: Union[Tensor, MutableMapping],
        y: Tensor,
        anchor_dict: Dict[str, Tensor],
        params_dict: Dict[str, Tensor],
    ) -> Tensor:
        """Compute the loss given a mini-batch under a linearized NN around anchor.

        # noqa: DAR101

        Returns:
            f(X, θ₀) + (J_θ₀ f(X, θ₀)) @ (θ - θ₀) with f the neural network, θ₀ the anchor
            point of the linearization, and θ the evaluation point.
        """
        output = linearized_model(anchor_dict, params_dict, X)
        return functional_call(loss_func, {}, (output, y))

    params_argnum = 3
    ggn_fn = hessian(linearized_loss, argnums=params_argnum)

    anchor_dict = {n: p.clone() for n, p in params_dict.items()}

    return blocks_to_matrix(ggn_fn(X, y, anchor_dict, params_dict))


def functorch_gradient(
    model_func: Module,
    loss_func: Module,
    params: List[Tensor],
    data: Iterable[Tuple[Union[Tensor, MutableMapping], Tensor]],
    input_key: Optional[str] = None,
) -> Tuple[Tensor]:
    """Compute the gradient with functorch.

    Args:
        model_func: A function that maps the mini-batch input X to predictions.
            Could be a PyTorch module representing a neural network.
        loss_func: Loss function criterion. Maps predictions and mini-batch labels
            to a scalar value.
        params: List of differentiable parameters used by the prediction function.
        data: Source from which mini-batches can be drawn, for instance a list of
            mini-batches ``[(X, y), ...]`` or a torch ``DataLoader``.
        input_key: Key to obtain the input tensor when ``X`` is a dict-like object.

    Returns:
        Gradient in same format as the parameters.
    """
    (dev,) = {p.device for p in params}
    X, y = _concatenate_batches(data, input_key, device=dev)
    params_dict = _make_params_dict(model_func, params)

    def loss(
        X: Union[Tensor, MutableMapping], y: Tensor, params_dict: Dict[str, Tensor]
    ) -> Tensor:
        """Compute the loss given a mini-batch and the neural network parameters.

        # noqa: DAR101
        # noqa: DAR201
        """
        output = functional_call(model_func, params_dict, X)
        return functional_call(loss_func, {}, (output, y))

    params_argnum = 2
    grad_fn = grad(loss, argnums=params_argnum)

    return tuple(grad_fn(X, y, params_dict).values())


def functorch_empirical_fisher(
    model_func: Module,
    loss_func: Module,
    params: List[Tensor],
    data: Iterable[Tuple[Union[Tensor, MutableMapping], Tensor]],
    input_key: Optional[str] = None,
) -> Tensor:
    """Compute the empirical Fisher with functorch.

    Args:
        model_func: A function that maps the mini-batch input X to predictions.
            Could be a PyTorch module representing a neural network.
        loss_func: Loss function criterion. Maps predictions and mini-batch labels
            to a scalar value.
        params: List of differentiable parameters used by the prediction function.
        data: Source from which mini-batches can be drawn, for instance a list of
            mini-batches ``[(X, y), ...]`` or a torch ``DataLoader``.
        input_key: Key to obtain the input tensor when ``X`` is a dict-like object.

    Returns:
        Square matrix containing the empirical Fisher.
    """
    (dev,) = {p.device for p in params}
    X, y = _concatenate_batches(data, input_key, device=dev)
    params_dict = _make_params_dict(model_func, params)

    def losses(
        X: Union[Tensor, MutableMapping], y: Tensor, params_dict: Dict[str, Tensor]
    ) -> Tensor:
        """Compute all elementary losses without reduction.

        An elementary loss results from a scalar entry of `y`.

        Args:
            X: Mini-batch input.
            y: Mini-batch labels.
            params_dict: Dictionary of parameters.

        Returns:
            1d tensor containing all elementary losses.
        """
        output = functional_call(model_func, params_dict, X)

        flatten_output = {
            MSELoss: "batch ... d_out -> (batch ... d_out)",
            BCEWithLogitsLoss: "batch ... d_out -> (batch ... d_out)",
            CrossEntropyLoss: "batch c ... -> (batch ...) c",
        }[loss_func.__class__]
        flatten_y = "... -> (...)"
        output_flat, y_flat = rearrange(output, flatten_output), rearrange(y, flatten_y)

        return stack(
            [
                functional_call(loss_func, {}, (o_el, y_el))
                for o_el, y_el in zip(output_flat, y_flat)
            ]
        )

    params_argnum = 2
    jac = jacrev(losses, argnums=params_argnum)(X, y, params_dict)
    jac = cat([j.flatten(start_dim=1) for j in jac.values()], dim=1)

    # the losses over which the expectation is taken
    num_losses = {
        CrossEntropyLoss: y.numel(),
        MSELoss: y.shape[:-1].numel(),
        BCEWithLogitsLoss: y.shape[:-1].numel(),
    }[loss_func.__class__]
    num_params = sum(p.numel() for p in params)

    # the losses which model the same random variable
    grouped_losses = y.numel() // num_losses
    jac = jac.reshape(num_losses, grouped_losses, num_params).sum(1)
    if (
        isinstance(loss_func, (MSELoss, BCEWithLogitsLoss))
        and loss_func.reduction == "mean"
    ):
        jac /= sqrt(grouped_losses)

    normalization = {"sum": 1, "mean": num_losses}[loss_func.reduction]
    return jac.T @ jac / normalization


def functorch_jacobian(
    model_func: Module,
    params: List[Tensor],
    data: Iterable[Tuple[Union[Tensor, MutableMapping], Tensor]],
    input_key: Optional[str] = None,
) -> Tensor:
    """Compute the Jacobian with functorch.

    Args:
        model_func: A function that maps the mini-batch input X to predictions.
            Could be a PyTorch module representing a neural network.
        params: List of differentiable parameters used by the prediction function.
        data: Source from which mini-batches can be drawn, for instance a list of
            mini-batches ``[(X, y), ...]`` or a torch ``DataLoader``.
        input_key: Key to obtain the input tensor when ``X`` is a dict-like object.

    Returns:
        Matrix containing the Jacobian. Has shape ``[N * C, D]`` where ``D`` is the
        total number of parameters, ``N`` the total number of data points, and ``C``
        the model's output space dimension.
    """
    (dev,) = {p.device for p in params}
    X, _ = _concatenate_batches(data, input_key, device=dev)
    params_dict = _make_params_dict(model_func, params)

    def model_fn_params_only(params_dict: Dict[str, Tensor]) -> Tensor:
        return functional_call(model_func, params_dict, X)

    # concatenate over flattened parameters and flattened outputs
    jac = jacrev(model_fn_params_only)(params_dict)
    jac = [
        j.flatten(start_dim=-p.dim())
        for j, p in zip(jac.values(), params_dict.values())
    ]
    jac = cat(jac, dim=-1).flatten(end_dim=-2)

    return jac


def _concatenate_batches(
    data: Iterable[Tuple[Union[Tensor, MutableMapping], Tensor]],
    input_key: Optional[str] = None,
    device: Optional[torch.device] = None,
) -> Tuple[Tensor, Tensor]:
    """Concatenate all batches in the dataset along the batch dimension.

    Args:
        data: A dataloader or iterable of batches.
        input_key: Key to obtain the input tensor when ``X`` is a dict-like object.
        device: The device the data should live in.

    Returns:
        Concatenated model inputs.
        Concatenated targets.

    Raises:
        ValueError: If ``X`` in ``data`` is a dict-like object and ``input_key`` is
            not provided.
    """
    X, y = list(zip(*list(data)))
    device = y[0].device if device is None else device
    y = cat(y).to(device)

    if isinstance(X[0], MutableMapping) and input_key is None:
        raise ValueError("input_key must be provided for dict-like X!")

    if isinstance(X[0], Tensor):
        return cat(X).to(device), y
    else:
        X = {input_key: cat([d[input_key] for d in X]).to(device)}
        return X, y


def _make_params_dict(model_func: Module, params: List[Tensor]) -> Dict[str, Tensor]:
    """Create a named dictionary for the parameter list.

    Required for ``functorch``'s ``functional_call`` API.

    Args:
        model_func: A PyTorch module representing a neural network.
        params: List of differentiable parameters used by the prediction function.

    Returns:
        Dictionary mapping parameter names to parameter tensors.
    """
    name_dict = {p.data_ptr(): name for name, p in model_func.named_parameters()}
    return {name_dict[p.data_ptr()]: p for p in params}



"""Contains tests for ``curvlinops/ggn``."""

from collections.abc import MutableMapping

from numpy import random
from pytest import raises

from curvlinops import GGNLinearOperator
from curvlinops.examples.functorch import functorch_ggn
from curvlinops.examples.utils import report_nonclose


def test_GGNLinearOperator_matvec(case, adjoint: bool):
    model_func, loss_func, params, data, batch_size_fn = case

    # Test when X is dict-like but batch_size_fn = None (default)
    if isinstance(data[0][0], MutableMapping):
        with raises(ValueError):
            op = GGNLinearOperator(model_func, loss_func, params, data)

    op = GGNLinearOperator(
        model_func, loss_func, params, data, batch_size_fn=batch_size_fn
    )
    op_functorch = (
        functorch_ggn(model_func, loss_func, params, data, input_key="x")
        .detach()
        .cpu()
        .numpy()
    )
    if adjoint:
        op, op_functorch = op.adjoint(), op_functorch.conj().T

    x = random.rand(op.shape[1])
    report_nonclose(op @ x, op_functorch @ x)


def test_GGNLinearOperator_matmat(case, adjoint: bool, num_vecs: int = 3):
    model_func, loss_func, params, data, batch_size_fn = case

    op = GGNLinearOperator(
        model_func, loss_func, params, data, batch_size_fn=batch_size_fn
    )
    op_functorch = (
        functorch_ggn(model_func, loss_func, params, data, input_key="x")
        .detach()
        .cpu()
        .numpy()
    )
    if adjoint:
        op, op_functorch = op.adjoint(), op_functorch.conj().T

    X = random.rand(op.shape[1], num_vecs)
    report_nonclose(op @ X, op_functorch @ X)



"""Test ``curvlinops.trace``."""

from typing import Union

from numpy import isclose, mean

from curvlinops.trace.hutchinson import HutchinsonTraceEstimator
from curvlinops.trace.meyer2020hutch import HutchPPTraceEstimator

DISTRIBUTIONS = ["rademacher", "normal"]
DISTRIBUTION_IDS = [f"distribution={distribution}" for distribution in DISTRIBUTIONS]


def _test_convergence(
    estimator: Union[HutchinsonTraceEstimator, HutchPPTraceEstimator],
    tr_A: float,
    distribution: str,
    max_samples: int = 20_000,
    chunk_size: int = 2_000,
    atol: float = 1e-3,
    rtol: float = 1e-2,
):
    """Test whether a trace estimator converges to the true trace.

    Args:
        estimator: The trace estimator.
        tr_A: True trace of the linear operator.
        distribution: Distribution of the random vectors used for the trace estimation.
        max_samples: Maximum number of samples to draw before error-ing.
            Default: ``20_000``
        chunk_size: Number of samples to draw before comparing against the true trace.
            Default: ``2_000``.
        atol: Absolute toleranc e used to compare with the exact trace.
            Default: ``1e-3``.
        rtol: Relative tolerance used to compare with the exact trace.
            Default: ``1e-2``.
    """
    samples = []

    while len(samples) < max_samples:
        samples.extend(
            [estimator.sample(distribution=distribution) for _ in range(chunk_size)]
        )
        tr_estimator = mean(samples)
        if isclose(tr_A, tr_estimator, atol=atol, rtol=rtol):
            # quit once the estimator has converged
            break

        print(f"{len(samples)} samples: Tr(A)={tr_A:.5f}≠{tr_estimator:.5f}.")

    tr_estimator = mean(samples)
    assert isclose(tr_A, tr_estimator, atol=atol, rtol=rtol)



"""Test ``curvlinops.diagonal.hutchinson``."""

from numpy import allclose, diag, isclose, mean
from numpy.random import rand, seed
from pytest import mark

from curvlinops import HutchinsonDiagonalEstimator

DISTRIBUTIONS = ["rademacher", "normal"]
DISTRIBUTION_IDS = [f"distribution={distribution}" for distribution in DISTRIBUTIONS]


@mark.parametrize("distribution", DISTRIBUTIONS, ids=DISTRIBUTION_IDS)
def test_HutchinsonDiagonalEstimator(distribution: str):
    """Test whether Hutchinson's diagonal estimator converges to the true diagonal.

    Args:
        distribution: Distribution of the random vectors used for the trace estimation.
    """
    seed(0)
    A = rand(10, 10)
    diag_A = diag(A)
    estimator = HutchinsonDiagonalEstimator(A)

    samples = []
    max_samples = 100_000
    chunk_size = 10_000
    atol, rtol = 1e-2, 5e-2

    while len(samples) < max_samples:
        samples.extend(
            [estimator.sample(distribution=distribution) for _ in range(chunk_size)]
        )
        diag_estimator = mean(samples, axis=0)
        if allclose(diag_A, diag_estimator, atol=atol, rtol=rtol):
            # quit once the estimator has converged
            break

        print(f"{len(samples)} samples:")
        for idx, (d1, d2) in enumerate(zip(diag_A, diag_estimator)):
            relation = "=" if isclose(d1, d2, atol=atol, rtol=rtol) else "≠"
            print(f"\tdiag(A)[{idx}]={d1:.5f}{relation}{d2:.5f}.")

    diag_estimator = mean(samples, axis=0)
    assert allclose(diag_A, diag_estimator, atol=atol, rtol=rtol)



"""Utility functions related to KFAC."""

from math import sqrt
from typing import Tuple, Union

from einconv import index_pattern
from einconv.utils import get_conv_paddings
from einops import einsum, rearrange, reduce
from torch import Tensor, diag, eye
from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, MSELoss
from torch.nn.functional import unfold
from torch.nn.modules.utils import _pair


def loss_hessian_matrix_sqrt(
    output_one_datum: Tensor,
    target_one_datum: Tensor,
    loss_func: Union[MSELoss, CrossEntropyLoss, BCEWithLogitsLoss],
) -> Tensor:
    r"""Compute the loss function's matrix square root for a sample's output.

    Args:
        output_one_datum: The model's prediction on a single datum. Has shape
            ``[1, C]`` where ``C`` is the number of classes (outputs of the neural
            network).
        target_one_datum: The label of the single datum.
        loss_func: The loss function.

    Returns:
        The matrix square root
        :math:`\mathbf{S}` of the Hessian. Has shape
        ``[C, C]`` and satisfies the relation

        .. math::
            \mathbf{S} \mathbf{S}^\top
            =
            \nabla^2_{\mathbf{f}} \ell(\mathbf{f}, \mathbf{y})
            \in \mathbb{R}^{C \times C}

        where :math:`\mathbf{f} := f(\mathbf{x}) \in \mathbb{R}^C` is the model's
        prediction on a single datum :math:`\mathbf{x}` and :math:`\mathbf{y}` is
        the label.

    Note:
        For :class:`torch.nn.MSELoss` (with :math:`c = 1` for ``reduction='sum'``
        and :math:`c = 1/C` for ``reduction='mean'``), we have:

        .. math::
            \ell(\mathbf{f}) &= c \sum_{i=1}^C (f_i - y_i)^2
            \\
            \nabla^2_{\mathbf{f}} \ell(\mathbf{f}, \mathbf{y}) &= 2 c \mathbf{I}_C
            \\
            \mathbf{S} &= \sqrt{2 c} \mathbf{I}_C

    Note:
        For :class:`torch.nn.CrossEntropyLoss` (with :math:`c = 1` irrespective of the
        reduction, :math:`\mathbf{p}:=\mathrm{softmax}(\mathbf{f}) \in \mathbb{R}^C`,
        and the element-wise natural logarithm :math:`\log`) we have:

        .. math::
            \ell(\mathbf{f}, y) = - c \log(\mathbf{p})^\top \mathrm{onehot}(y)
            \\
            \nabla^2_{\mathbf{f}} \ell(\mathbf{f}, y)
            =
            c \left(
            \mathrm{diag}(\mathbf{p}) - \mathbf{p} \mathbf{p}^\top
            \right)
            \\
            \mathbf{S} = \sqrt{c} \left(
            \mathrm{diag}(\sqrt{\mathbf{p}}) - \sqrt{\mathbf{p}} \mathbf{p}^\top
            \right)\,,

       where the square root is applied element-wise. See for instance Example 5.1 of
       `this thesis <https://d-nb.info/1280233206/34>`_ or equations (5) and (6) of
       `this paper <https://arxiv.org/abs/1901.08244>`_.

    Note:
        For :class:`torch.nn.BCEWithLogitsLoss` (with :math:`c = 1` for ``reduction='sum'``
        and :math:`c = 1/C` for ``reduction='mean'``) we have (:math:`\sigma` is the sigmoid
        function, and assuming binary labels):

        .. math::
            \ell(\mathbf{f})
            &=
            c \sum_{i=1}^C - y_i \log(\sigma(f_i)) - (1 - y_i) \log(1 - \sigma(f_i))
            \\
            \nabla^2_{\mathbf{f}} \ell(\mathbf{f}, \mathbf{y})
            &=
            c \mathrm{diag}( \sigma(f_i) \odot (1 - \sigma(f_i)) )
            \\
            \mathbf{S}
            &=
            \sqrt{c} \mathrm{diag}(\sqrt{\sigma(f_i) \odot (1 - \sigma(f_i))})\,,

        where the square root is applied element-wise.

    Raises:
        ValueError: If the batch size is not one, or the output is not 2d.
        NotImplementedError: If the loss function is not supported.
        NotImplementedError: If the loss function is ``BCEWithLogitsLoss`` but the
            target is not binary.
    """
    if output_one_datum.ndim != 2 or output_one_datum.shape[0] != 1:
        raise ValueError(
            f"Expected 'output_one_datum' to be 2d with shape [1, C], got "
            f"{output_one_datum.shape}"
        )
    if target_one_datum.shape[0] != 1:  # targets for 2d predictions are sometimes 1d
        raise ValueError(
            "Expected 'target_one_datum' to have batch_size 1."
            + f" Got {target_one_datum.shape}."
        )
    output = output_one_datum.squeeze(0)
    output_dim = output.numel()

    if isinstance(loss_func, MSELoss):
        c = {"sum": 1.0, "mean": 1.0 / output_dim}[loss_func.reduction]
        return eye(output_dim, device=output.device, dtype=output.dtype).mul_(
            sqrt(2 * c)
        )

    elif isinstance(loss_func, CrossEntropyLoss):
        c = 1.0
        p = output_one_datum.softmax(dim=1).squeeze()
        p_sqrt = p.sqrt()
        return (diag(p_sqrt) - einsum(p, p_sqrt, "i, j -> i j")).mul_(sqrt(c))

    elif isinstance(loss_func, BCEWithLogitsLoss):
        unique = set(target_one_datum.unique().flatten().tolist())
        if not unique.issubset({0, 1}):
            raise NotImplementedError(
                "Only binary targets (0, 1) are currently supported with"
                + f"BCEWithLogitsLoss. Got {unique}."
            )

        c = {"sum": 1.0, "mean": 1.0 / output_dim}[loss_func.reduction]
        p = output_one_datum.sigmoid().squeeze(0)
        hess_diag = sqrt(c) * (p * (1 - p)).sqrt()
        return hess_diag.diag()
    else:
        raise NotImplementedError(f"Loss function {loss_func} not supported.")


def extract_patches(
    x: Tensor,
    kernel_size: Union[Tuple[int, int], int],
    stride: Union[Tuple[int, int], int],
    padding: Union[Tuple[int, int], int, str],
    dilation: Union[Tuple[int, int], int],
    groups: int,
) -> Tensor:
    """Extract patches from the input of a 2d-convolution.

    The patches are averaged over channel groups.

    Args:
        x: Input to a 2d-convolution. Has shape ``[batch_size, C_in, I1, I2]``.
        kernel_size: The convolution's kernel size supplied as 2-tuple or integer.
        stride: The convolution's stride supplied as 2-tuple or integer.
        padding: The convolution's padding supplied as 2-tuple, integer, or string.
        dilation: The convolution's dilation supplied as 2-tuple or integer.
        groups: The number of channel groups.

    Returns:
        A tensor of shape ``[batch_size, O1 * O2, C_in // groups * K1 * K2]`` where
        each column ``[b, o1_o2, :]`` contains the flattened patch of sample ``b`` used
        for output location ``(o1, o2)``, averaged over channel groups.

    Raises:
        NotImplementedError: If ``padding`` is a string that would lead to unequal
            padding along a dimension.
    """
    if isinstance(padding, str):  # get padding as integers
        padding_as_int = []
        for k, s, d in zip(_pair(kernel_size), _pair(stride), _pair(dilation)):
            p_left, p_right = get_conv_paddings(k, s, padding, d)
            if p_left != p_right:
                raise NotImplementedError("Unequal padding not supported in unfold.")
            padding_as_int.append(p_left)
        padding = tuple(padding_as_int)

    # average channel groups
    x = rearrange(x, "b (g c_in) i1 i2 -> b g c_in i1 i2", g=groups)
    x = reduce(x, "b g c_in i1 i2 -> b c_in i1 i2", "mean")

    x_unfold = unfold(x, kernel_size, dilation=dilation, padding=padding, stride=stride)
    return rearrange(x_unfold, "b c_in_k1_k2 o1_o2 -> b o1_o2 c_in_k1_k2")


def extract_averaged_patches(
    x: Tensor,
    kernel_size: Union[Tuple[int, int], int],
    stride: Union[Tuple[int, int], int],
    padding: Union[Tuple[int, int], int, str],
    dilation: Union[Tuple[int, int], int],
    groups: int,
) -> Tensor:
    """Extract averaged patches from the input of a 2d-convolution.

    The patches are averaged over channel groups and output locations.

    Uses the tensor network formulation of convolution from
    `Dangel, 2023 <https://arxiv.org/abs/2307.02275>`_.

    Args:
        x: Input to a 2d-convolution. Has shape ``[batch_size, C_in, I1, I2]``.
        kernel_size: The convolution's kernel size supplied as 2-tuple or integer.
        stride: The convolution's stride supplied as 2-tuple or integer.
        padding: The convolution's padding supplied as 2-tuple, integer, or string.
        dilation: The convolution's dilation supplied as 2-tuple or integer.
        groups: The number of channel groups.

    Returns:
        A tensor of shape ``[batch_size, C_in // groups * K1 * K2]`` where each column
        ``[b, :]`` contains the flattened patch of sample ``b`` averaged over all output
        locations and channel groups.
    """
    # average channel groups
    x = rearrange(x, "b (g c_in) i1 i2 -> b g c_in i1 i2", g=groups)
    x = reduce(x, "b g c_in i1 i2 -> b c_in i1 i2", "mean")

    # TODO For convolutions with special structure, we don't even need to compute
    # the index pattern tensors, or can resort to contracting only slices thereof.
    # In order for this to work `einconv`'s TN simplification mechanism must first
    # be refactored to work purely symbolically. Once this is done, it will be
    # possible to do the below even more efficiently (memory and run time) for
    # structured convolutions.

    # compute index pattern tensors, average output dimension
    patterns = []
    input_sizes = x.shape[-2:]
    for i, k, s, p, d in zip(
        input_sizes,
        _pair(kernel_size),
        _pair(stride),
        (padding, padding) if isinstance(padding, str) else _pair(padding),
        _pair(dilation),
    ):
        pi = index_pattern(
            i, k, stride=s, padding=p, dilation=d, dtype=x.dtype, device=x.device
        )
        pi = reduce(pi, "k o i -> k i", "mean")
        patterns.append(pi)

    x = einsum(x, *patterns, "b c_in i1 i2, k1 i1, k2 i2 -> b c_in k1 k2")
    return rearrange(x, "b c_in k1 k2 -> b (c_in k1 k2)")



"""Contains tests for ``curvlinops/submatrix`` with random matrices."""

from typing import List, Tuple

from numpy import eye, ndarray, random
from pytest import fixture, mark, raises
from scipy.sparse.linalg import aslinearoperator

from curvlinops.examples.utils import report_nonclose
from curvlinops.submatrix import SubmatrixLinearOperator

SUBMATRIX_CASES = [
    # same indices
    {
        "A_fn": lambda: random.rand(300, 300),
        "row_idxs_fn": lambda: [0, 13, 42, 299],
        "col_idxs_fn": lambda: [0, 13, 42, 299],
        "seed": 0,
    },
    # different indices
    {
        "A_fn": lambda: random.rand(200, 250),
        "row_idxs_fn": lambda: [5, 67, 128],
        "col_idxs_fn": lambda: [83, 21],
        "seed": 0,
    },
]


@fixture(params=SUBMATRIX_CASES)
def submatrix_case(request) -> Tuple[ndarray, List[int], List[int]]:
    case = request.param
    random.seed(case["seed"])
    return case["A_fn"](), case["row_idxs_fn"](), case["col_idxs_fn"]()


@mark.parametrize("adjoint", [False, True], ids=["", "adjoint"])
def test_SubmatrixLinearOperator__matvec(
    submatrix_case: Tuple[ndarray, List[int], List[int]], adjoint: bool
):
    """Test the matrix-vector multiplication of a submatrix linear operator.

    Args:
        submatrix_case: A tuple with a random matrix and two index lists.
        adjoint: Whether to take the operator's adjoint before multiplying.
    """
    A, row_idxs, col_idxs = submatrix_case

    A_sub = A[row_idxs, :][:, col_idxs]
    A_sub_linop = SubmatrixLinearOperator(aslinearoperator(A), row_idxs, col_idxs)

    if adjoint:
        A_sub = A_sub.conj().T
        A_sub_linop = A_sub_linop.adjoint()

    x = random.rand(A_sub.shape[1])
    A_sub_linop_x = A_sub_linop @ x

    assert A_sub_linop_x.shape == ((len(col_idxs),) if adjoint else (len(row_idxs),))
    report_nonclose(A_sub @ x, A_sub_linop_x)


@mark.parametrize("adjoint", [False, True], ids=["", "adjoint"])
def test_SubmatrixLinearOperator__matmat(
    submatrix_case: Tuple[ndarray, List[int], List[int]],
    adjoint: bool,
    num_vecs: int = 3,
):
    """Test the matrix-matrix multiplication of a submatrix linear operator.

    Args:
        submatrix_case: A tuple with a random matrix and two index lists.
        adjoint: Whether to take the operator's adjoint before multiplying.
        num_vecs: The number of vectors to multiply. Default: ``3``.
    """
    A, row_idxs, col_idxs = submatrix_case

    A_sub = A[row_idxs, :][:, col_idxs]
    A_sub_linop = SubmatrixLinearOperator(aslinearoperator(A), row_idxs, col_idxs)

    if adjoint:
        A_sub = A_sub.conj().T
        A_sub_linop = A_sub_linop.adjoint()

    X = random.rand(A_sub.shape[1], num_vecs)
    A_sub_linop_X = A_sub_linop @ X

    assert A_sub_linop_X.shape == (
        (len(col_idxs), num_vecs) if adjoint else (len(row_idxs), num_vecs)
    )
    report_nonclose(A_sub @ X, A_sub_linop_X)


def test_SubmatrixLinearOperator_set_submatrix():
    A = aslinearoperator(eye(10))

    invalid_idxs = [
        [[0.0], [0]],  # wrong type in row_idxs
        [[0], [0.0]],  # wrong type in col_idxs
        [[2, 1, 2], [3]],  # duplicate entries in row_idxs
        [[3], [2, 1, 2]],  # duplicate entries in col_idxs
        [[10, 5], [2]],  # out-of-bounds in row_idxs
        [[6, 5], [-1]],  # out-of-bounds in col_idxs
    ]

    for row_idxs, col_idxs in invalid_idxs:
        with raises(ValueError):
            SubmatrixLinearOperator(A, row_idxs, col_idxs)



"""Implements linear operator inverses."""

from math import sqrt
from typing import Any, Callable, Dict, Optional, Tuple, TypeVar, Union
from warnings import warn

from einops import einsum, rearrange
from numpy import allclose, column_stack, ndarray
from scipy.sparse.linalg import LinearOperator, cg, lsmr
from torch import Tensor, cat, cholesky_inverse, eye, float64, outer
from torch.linalg import cholesky, eigh

from curvlinops.kfac import KFACLinearOperator, ParameterMatrixType

KFACInvType = TypeVar(
    "KFACInvType", Optional[Tensor], Tuple[Optional[Tensor], Optional[Tensor]]
)


class _InverseLinearOperator(LinearOperator):
    """Base class for (approximate) inverses of linear operators."""

    def _matmat(self, X: ndarray) -> ndarray:
        """Matrix-matrix multiplication.

        Args:
            X: Matrix for multiplication.

        Returns:
            Matrix-multiplication result ``A⁻¹@ X``.
        """
        return column_stack([self @ col for col in X.T])


class CGInverseLinearOperator(_InverseLinearOperator):
    """Class for inverse linear operators via conjugate gradients."""

    def __init__(self, A: LinearOperator):
        """Store the linear operator whose inverse should be represented.

        Args:
            A: Linear operator whose inverse is formed. Must be symmetric and
                positive-definite.
        """
        super().__init__(A.dtype, A.shape)
        self._A = A

        # CG hyperparameters
        self.set_cg_hyperparameters()

    def set_cg_hyperparameters(
        self,
        x0: Optional[ndarray] = None,
        maxiter: Optional[int] = None,
        M: Optional[Union[ndarray, LinearOperator]] = None,
        callback: Optional[Callable] = None,
        atol: Optional[float] = None,
        tol: Optional[float] = 1e-5,
    ):
        """Store hyperparameters for CG.

        They will be used to approximate the inverse matrix-vector products.

        For more detail, see
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.cg.html.

        # noqa: DAR101
        """
        self._cg_hyperparameters = {
            "x0": x0,
            "maxiter": maxiter,
            "M": M,
            "callback": callback,
            "atol": atol,
            "tol": tol,
        }

    def _matvec(self, x: ndarray) -> ndarray:
        """Multiply x by the inverse of A.

        Args:
             x: Vector for multiplication.

        Returns:
             Result of inverse matrix-vector multiplication, ``A⁻¹ @ x``.
        """
        result, _ = cg(self._A, x, **self._cg_hyperparameters)
        return result


class LSMRInverseLinearOperator(_InverseLinearOperator):
    """Class for inverse linear operators via LSMR.

    See https://arxiv.org/abs/1006.0758 for details on the LSMR algorithm.
    """

    def __init__(self, A: LinearOperator):
        """Store the linear operator whose inverse should be represented.

        Args:
            A: Linear operator whose inverse is formed.
        """
        super().__init__(A.dtype, A.shape)
        self._A = A

        # LSMR hyperparameters
        self.set_lsmr_hyperparameters()

    def set_lsmr_hyperparameters(
        self,
        damp: float = 0.0,
        atol: Optional[float] = 1e-6,
        btol: Optional[float] = 1e-6,
        conlim: Optional[float] = 1e-8,
        maxiter: Optional[int] = None,
        show: Optional[bool] = False,
        x0: Optional[ndarray] = None,
    ):
        """Store hyperparameters for LSMR.

        They will be used to approximate the inverse matrix-vector products.

        For more detail, see
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.lsmr.html.

        # noqa: DAR101
        """
        self._lsmr_hyperparameters = {
            "damp": damp,
            "atol": atol,
            "btol": btol,
            "conlim": conlim,
            "maxiter": maxiter,
            "show": show,
            "x0": x0,
        }

    def matvec_with_info(
        self, x: ndarray
    ) -> Tuple[ndarray, int, int, float, float, float, float, float]:
        """Multiply x by the inverse of A and return additional information.

        Args:
             x: Vector for multiplication.

        Returns:
            Result of inverse matrix-vector multiplication, ``A⁻¹ @ x`` with additional
            information; see
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.lsmr.html
            for details (same return values).
        """
        return lsmr(self._A, x, **self._lsmr_hyperparameters)

    def _matvec(self, x: ndarray) -> ndarray:
        """Multiply x by the inverse of A.

        Args:
             x: Vector for multiplication.

        Returns:
             Result of inverse matrix-vector multiplication, ``A⁻¹ @ x``.
        """
        return self.matvec_with_info(x)[0]


class NeumannInverseLinearOperator(_InverseLinearOperator):
    """Class for inverse linear operators via truncated Neumann series.

    # noqa: B950

    See https://en.wikipedia.org/w/index.php?title=Neumann_series&oldid=1131424698#Approximate_matrix_inversion.

    Motivated by (referred to as lorraine2020optimizing in the following)

    - Lorraine, J., Vicol, P., & Duvenaud, D. (2020). Optimizing millions of
      hyperparameters by implicit differentiation. In International Conference on
      Artificial Intelligence and Statistics (AISTATS).

    .. warning::
        The Neumann series can be non-convergent. In this case, the iterations
        will become numerically unstable, leading to ``NaN``s.

    .. warning::
        The Neumann series can converge slowly.
        Use :py:class:`curvlinops.CGInverLinearOperator` for better accuracy.
    """

    DEFAULT_NUM_TERMS = 100
    DEFAULT_SCALE = 1.0
    DEFAULT_CHECK_NAN = True

    def __init__(
        self,
        A: LinearOperator,
        num_terms: int = DEFAULT_NUM_TERMS,
        scale: float = DEFAULT_SCALE,
        check_nan: bool = DEFAULT_CHECK_NAN,
    ):
        r"""Store the linear operator whose inverse should be represented.

        The Neumann series for an invertible linear operator :math:`\mathbf{A}` is

        .. math::
            \mathbf{A}^{-1}
            =
            \sum_{k=0}^{\infty}
            \left(\mathbf{I} - \mathbf{A} \right)^k\,.

        and is convergent if all eigenvalues satisfy
        :math:`0 < \lambda(\mathbf{A}) < 2`.

        By re-rescaling the matrix by ``scale`` (:math:`\alpha`), we have:

        .. math::
            \mathbf{A}^{-1}
            =
            \alpha (\alpha \mathbf{A})^{-1}
            =
            \alpha \sum_{k=0}^{\infty}
            \left(\mathbf{I} - \alpha \mathbf{A} \right)^k\,,

        which and is convergent if :math:`0 < \lambda(\mathbf{A}) < \frac{2}{\alpha}`.

        Additionally, we truncate the series at ``num_terms`` (:math:`K`):

        .. math::
            \mathbf{A}^{-1}
            \approx
            \alpha \sum_{k=0}^{K}
            \left(\mathbf{I} - \alpha \mathbf{A} \right)^k\,.

        Args:
            A: Linear operator whose inverse is formed.
            num_terms: Number of terms in the truncated Neumann series.
                Default: ``100``.
            scale: Scale applied to the matrix in the Neumann iteration. Crucial
                for convergence of Neumann series (details above). Default: ``1.0``.
            check_nan: Whether to check for NaNs while applying the truncated Neumann
                series. Default: ``True``.
        """
        super().__init__(A.dtype, A.shape)
        self._A = A
        self.set_neumann_hyperparameters(
            num_terms=num_terms, scale=scale, check_nan=check_nan
        )

    def set_neumann_hyperparameters(
        self,
        num_terms: int = DEFAULT_NUM_TERMS,
        scale: float = DEFAULT_SCALE,
        check_nan: bool = DEFAULT_CHECK_NAN,
    ):
        """Store hyperparameters for the truncated Neumann series.

        Args:
            num_terms: Number of terms in the truncated Neumann series.
                Default: ``100``.
            scale: Scale applied to the matrix in the Neumann iteration. Crucial
                for convergence of Neumann series. Default: ``1.0``.
            check_nan: Whether to check for NaNs while applying the truncated Neumann
                series. Default: ``True``.
        """
        self._num_terms = num_terms
        self._scale = scale
        self._check_nan = check_nan

    def _matvec(self, x: ndarray) -> ndarray:
        """Multiply x by the inverse of A.

        Args:
             x: Vector for multiplication.

        Returns:
             Result of inverse matrix-vector multiplication, ``A⁻¹ @ x``.

        Raises:
            ValueError: If ``NaN`` check is turned on and ``NaN``s are detected.
        """
        result, v = x.copy(), x.copy()

        for idx in range(self._num_terms):
            v = v - self._scale * (self._A @ v)
            result = result + v

            if self._check_nan and not allclose(result, result):
                raise ValueError(
                    f"Detected NaNs after application of {idx}-th term."
                    + " This is probably because the Neumann series is non-convergent."
                    + " Try decreasing `scale` and read the comment on convergence."
                )

        return self._scale * result


class KFACInverseLinearOperator(_InverseLinearOperator):
    """Class to invert instances of the ``KFACLinearOperator``."""

    def __init__(
        self,
        A: KFACLinearOperator,
        damping: Union[float, Tuple[float, float]] = 0.0,
        use_heuristic_damping: bool = False,
        min_damping: float = 1e-8,
        use_exact_damping: bool = False,
        cache: bool = True,
        retry_double_precision: bool = True,
    ):
        r"""Store the linear operator whose inverse should be represented.

        Args:
            A: ``KFACLinearOperator`` whose inverse is formed.
            damping: Damping value(s) for all input and gradient covariances. If tuple,
                the first value is used for the input covariances and the second value
                for the gradient covariances. Note that if heuristic or exact damping is
                used the damping cannot be a tuple. Default: ``0.``.
            use_heuristic_damping: Whether to use a heuristic damping strategy by
                `Martens and Grosse, 2015 <https://arxiv.org/abs/1503.05671>`_
                (Section 6.3). For input covariances
                :math:`A \in \mathbb{R}^{n \times n}` and gradient covariances
                :math:`B \in \mathbb{R}^{m \times m}`, we define
                :math:`\pi := \sqrt{\frac{m\; \text{tr}(A)}{n\; \text{tr}(B)}}` and set the
                damping for the input covariances :math:`A` to
                :math:`\max(\pi\; \sqrt{\text{damping}}, \text{min_damping})` and for the
                gradient covariances :math:`B` to
                :math:`\max(\frac{1}{\pi}\; \sqrt{\text{damping}}, \text{min_damping})`.
                Default: ``False``.
            min_damping: Minimum damping value. Only used if ``use_heuristic_damping``
                is ``True``. Default: ``1e-8``.
            use_exact_damping: Whether to use exact damping, i.e. to invert
                :math:`(A \otimes B) + \text{damping}\; \mathbf{I}`. This is implemented
                via eigendecompositions of the Kronecker factors, e.g. see equation (21)
                in `Grosse et al., 2023 <https://arxiv.org/abs/2308.03296>`_.
                Note that the eigendecomposition synchronizes the device with the CPU.
                Default: ``False``.
            cache: Whether to cache the inverses of the Kronecker factors.
                Default: ``True``.
            retry_double_precision: Whether to retry Cholesky decomposition used for
                inversion in double precision. Default: ``True``.

        Raises:
            ValueError: If the linear operator is not a ``KFACLinearOperator``.
            ValueError: If both heuristic and exact damping are selected.
            ValueError: If heuristic or exact damping is used and the damping value is a
                tuple.
        """
        if not isinstance(A, KFACLinearOperator):
            raise ValueError(
                "The input `A` must be an instance of `KFACLinearOperator`."
            )
        super().__init__(A.dtype, A.shape)
        self._A = A
        if use_heuristic_damping and use_exact_damping:
            raise ValueError("Either use heuristic damping or exact damping, not both.")
        if (use_heuristic_damping or use_exact_damping) and isinstance(damping, tuple):
            raise ValueError(
                "Heuristic and exact damping require a single damping value."
            )

        self._damping = damping
        self._use_heuristic_damping = use_heuristic_damping
        self._min_damping = min_damping
        self._use_exact_damping = use_exact_damping
        self._cache = cache
        self._retry_double_precision = retry_double_precision
        self._inverse_input_covariances: Dict[str, KFACInvType] = {}
        self._inverse_gradient_covariances: Dict[str, KFACInvType] = {}

    def _compute_damping(
        self, aaT: Optional[Tensor], ggT: Optional[Tensor]
    ) -> Tuple[float, float]:
        """Compute the damping values for the input and gradient covariances.

        Args:
            aaT: Input covariance matrix. ``None`` for biases.
            ggT: Gradient covariance matrix.

        Returns:
            Damping values for the input and gradient covariances.
        """
        if self._use_heuristic_damping and aaT is not None and ggT is not None:
            # Martens and Grosse, 2015 (https://arxiv.org/abs/1503.05671) (Section 6.3)
            aaT_eig_mean = aaT.trace() / aaT.shape[0]
            ggT_eig_mean = ggT.trace() / ggT.shape[0]
            if aaT_eig_mean >= 0.0 and ggT_eig_mean > 0.0:
                sqrt_eig_mean_ratio = (aaT_eig_mean / ggT_eig_mean).sqrt()
                sqrt_damping = sqrt(self._damping)
                damping_aaT = max(sqrt_damping * sqrt_eig_mean_ratio, self._min_damping)
                damping_ggT = max(sqrt_damping / sqrt_eig_mean_ratio, self._min_damping)
                return damping_aaT, damping_ggT

        if isinstance(self._damping, tuple):
            return self._damping

        return self._damping, self._damping

    def _damped_cholesky(self, M: Tensor, damping: float) -> Tensor:
        """Compute the Cholesky decomposition of a matrix with damping.

        Args:
            M: Matrix for Cholesky decomposition.
            damping: Damping value.

        Returns:
            Cholesky decomposition of the matrix with damping.
        """
        return cholesky(
            M.add(eye(M.shape[0], dtype=M.dtype, device=M.device), alpha=damping)
        )

    def _compute_inverse_factors(
        self, aaT: Optional[Tensor], ggT: Optional[Tensor]
    ) -> Tuple[KFACInvType, KFACInvType]:
        """Compute the inverses of the Kronecker factors for a given layer.

        Args:
            aaT: Input covariance matrix. ``None`` for biases.
            ggT: Gradient covariance matrix.

        Returns:
            Tuple of inverses (or eigendecompositions) of the input and gradient
            covariance Kronecker factors. Can be ``None`` if the input or gradient
            covariance is ``None`` (e.g. the input covariances for biases).

        Raises:
            RuntimeError: If a Cholesky decomposition (and optionally the retry in
                double precision) fails.
        """
        if self._use_exact_damping:
            # Compute eigendecomposition to perform damped preconditioning in
            # Kronecker-factored eigenbasis (KFE).
            aaT_eigvals, aaT_eigvecs = (None, None) if aaT is None else eigh(aaT)
            ggT_eigvals, ggT_eigvecs = (None, None) if ggT is None else eigh(ggT)
            return (aaT_eigvecs, aaT_eigvals), (ggT_eigvecs, ggT_eigvals)
        else:
            damping_aaT, damping_ggT = self._compute_damping(aaT, ggT)

            # Compute inverse of aaT via Cholesky decomposition
            try:
                aaT_chol = (
                    None if aaT is None else self._damped_cholesky(aaT, damping_aaT)
                )
            except RuntimeError as error:
                if self._retry_double_precision and aaT.dtype != float64:
                    warn(
                        f"Failed to compute Cholesky decomposition in {aaT.dtype} "
                        f"precision with error {error}. "
                        "Retrying in double precision..."
                    )
                    # Retry in double precision
                    original_type = aaT.dtype
                    aaT = aaT.to(float64)
                    aaT_chol = self._damped_cholesky(aaT, damping_aaT)
                    aaT_chol = aaT_chol.to(original_type)
                else:
                    raise error
            aaT_inv = None if aaT_chol is None else cholesky_inverse(aaT_chol)

            # Compute inverse of ggT via Cholesky decomposition
            try:
                ggT_chol = (
                    None if ggT is None else self._damped_cholesky(ggT, damping_ggT)
                )
            except RuntimeError as error:
                if self._retry_double_precision and ggT.dtype != float64:
                    warn(
                        f"Failed to compute Cholesky decomposition in {ggT.dtype} "
                        f"precision with error {error}. "
                        "Retrying in double precision..."
                    )
                    # Retry in double precision
                    original_dtype = ggT.dtype
                    ggT = ggT.to(float64)
                    ggT_chol = self._damped_cholesky(ggT, damping_ggT)
                    ggT_chol = ggT_chol.to(original_dtype)
                else:
                    raise error
            ggT_inv = None if ggT_chol is None else cholesky_inverse(ggT_chol)

            return aaT_inv, ggT_inv

    def _compute_or_get_cached_inverse(
        self, name: str
    ) -> Tuple[KFACInvType, KFACInvType]:
        """Invert the Kronecker factors of the KFACLinearOperator or retrieve them.

        Args:
            name: Name of the layer for which to invert Kronecker factors.

        Returns:
            Tuple of inverses (or eigendecompositions) of the input and gradient
            covariance Kronecker factors. Can be ``None`` if the input or gradient
            covariance is ``None`` (e.g. the input covariances for biases).
        """
        if name in self._inverse_input_covariances:
            aaT_inv = self._inverse_input_covariances.get(name)
            ggT_inv = self._inverse_gradient_covariances.get(name)
            return aaT_inv, ggT_inv

        aaT = self._A._input_covariances.get(name)
        ggT = self._A._gradient_covariances.get(name)
        aaT_inv, ggT_inv = self._compute_inverse_factors(aaT, ggT)

        if self._cache:
            self._inverse_input_covariances[name] = aaT_inv
            self._inverse_gradient_covariances[name] = ggT_inv

        return aaT_inv, ggT_inv

    def _left_and_right_multiply(
        self, M_joint: Tensor, aaT_inv: KFACInvType, ggT_inv: KFACInvType
    ) -> Tensor:
        """Left and right multiply matrix with inverse Kronecker factors.

        Args:
            M_joint: Matrix for multiplication.
            aaT_inv: Inverse of the input covariance Kronecker factor. ``None`` for
                biases.
            ggT_inv: Inverse of the gradient covariance Kronecker factor.

        Returns:
            Matrix-multiplication result ``KFAC⁻¹ @ M_joint``.
        """
        if self._use_exact_damping:
            # Perform damped preconditioning in KFE, e.g. see equation (21) in
            # https://arxiv.org/abs/2308.03296.
            aaT_eigvecs, aaT_eigvals = aaT_inv
            ggT_eigvecs, ggT_eigvals = ggT_inv
            # Transform in eigenbasis.
            M_joint = einsum(
                ggT_eigvecs, M_joint, aaT_eigvecs, "i j, m i k, k l -> m j l"
            )
            # Divide by damped eigenvalues to perform the inversion.
            M_joint.div_(outer(ggT_eigvals, aaT_eigvals).add_(self._damping))
            # Transform back to standard basis.
            M_joint = einsum(
                ggT_eigvecs, M_joint, aaT_eigvecs, "i j, m j k, l k -> m i l"
            )
        else:
            M_joint = einsum(ggT_inv, M_joint, aaT_inv, "i j, m j k, k l -> m i l")
        return M_joint

    def _separate_left_and_right_multiply(
        self,
        M_torch: Tensor,
        param_pos: Dict[str, int],
        aaT_inv: KFACInvType,
        ggT_inv: KFACInvType,
    ) -> Tensor:
        """Multiply matrix with inverse Kronecker factors for separated weight and bias.

        Args:
            M_torch: Matrix for multiplication.
            param_pos: Dictionary with positions of the weight and bias parameters.
            aaT_inv: Inverse of the input covariance Kronecker factor. ``None`` for
                biases.
            ggT_inv: Inverse of the gradient covariance Kronecker factor.

        Returns:
            Matrix-multiplication result ``KFAC⁻¹ @ M_torch``.
        """
        if self._use_exact_damping:
            # Perform damped preconditioning in KFE, e.g. see equation (21) in
            # https://arxiv.org/abs/2308.03296.
            aaT_eigvecs, aaT_eigvals = aaT_inv
            ggT_eigvecs, ggT_eigvals = ggT_inv

        for p_name, pos in param_pos.items():
            # for weights we need to multiply from the right with aaT
            # for weights and biases we need to multiply from the left with ggT
            if p_name == "weight":
                M_w = rearrange(M_torch[pos], "m c_out ... -> m c_out (...)")
                aaT_fac = aaT_eigvecs if self._use_exact_damping else aaT_inv
                # If `use_exact_damping` is `True`, we transform to eigenbasis
                M_torch[pos] = einsum(M_w, aaT_fac, "m i j, j k -> m i k")

            ggT_fac = ggT_eigvecs if self._use_exact_damping else ggT_inv
            dims = (
                "m i ... -> m j ..."
                if self._use_exact_damping
                else " m j ... -> m i ..."
            )
            # If `use_exact_damping` is `True`, we transform to eigenbasis
            M_torch[pos] = einsum(ggT_fac, M_torch[pos], f"i j, {dims}")

            if self._use_exact_damping:
                # Divide by damped eigenvalues to perform the inversion and transform
                # back to standard basis.
                if p_name == "weight":
                    M_torch[pos].div_(
                        outer(ggT_eigvals, aaT_eigvals).add_(self._damping)
                    )
                    M_torch[pos] = einsum(
                        M_torch[pos], aaT_eigvecs, "m i j, k j -> m i k"
                    )
                else:
                    M_torch[pos].div_(ggT_eigvals.add_(self._damping))
                M_torch[pos] = einsum(
                    ggT_eigvecs, M_torch[pos], "i j, m j ... -> m i ..."
                )

        return M_torch

    def torch_matmat(self, M_torch: ParameterMatrixType) -> ParameterMatrixType:
        """Apply the inverse of KFAC to a matrix (multiple vectors) in PyTorch.

        This allows for matrix-matrix products with the inverse KFAC approximation in
        PyTorch without converting tensors to numpy arrays, which avoids unnecessary
        device transfers when working with GPUs and flattening/concatenating.

        Args:
            M_torch: Matrix for multiplication. If list of tensors, each entry has the
                same shape as a parameter with an additional leading dimension of size
                ``K`` for the columns, i.e. ``[(K,) + p1.shape), (K,) + p2.shape, ...]``.
                If tensor, has shape ``[D, K]`` with some ``K``.

        Returns:
            Matrix-multiplication result ``KFAC⁻¹ @ M``. Return type is the same as the
            type of the input. If list of tensors, each entry has the same shape as a
            parameter with an additional leading dimension of size ``K`` for the columns,
            i.e. ``[(K,) + p1.shape, (K,) + p2.shape, ...]``. If tensor, has shape
            ``[D, K]`` with some ``K``.
        """
        return_tensor, M_torch = self._A._check_input_type_and_preprocess(M_torch)
        if not self._A._input_covariances and not self._A._gradient_covariances:
            self._A._compute_kfac()

        for mod_name, param_pos in self._A._mapping.items():
            # retrieve the inverses of the Kronecker factors from cache or invert them
            aaT_inv, ggT_inv = self._compute_or_get_cached_inverse(mod_name)
            # cache the weight shape to ensure correct shapes are returned
            if "weight" in param_pos:
                weight_shape = M_torch[param_pos["weight"]].shape

            # bias and weights are treated jointly
            if (
                not self._A._separate_weight_and_bias
                and "weight" in param_pos.keys()
                and "bias" in param_pos.keys()
            ):
                w_pos, b_pos = param_pos["weight"], param_pos["bias"]
                M_w = rearrange(M_torch[w_pos], "m c_out ... -> m c_out (...)")
                M_joint = cat([M_w, M_torch[b_pos].unsqueeze(2)], dim=2)
                M_joint = self._left_and_right_multiply(M_joint, aaT_inv, ggT_inv)
                w_cols = M_w.shape[2]
                M_torch[w_pos], M_torch[b_pos] = M_joint.split([w_cols, 1], dim=2)
            else:
                M_torch = self._separate_left_and_right_multiply(
                    M_torch, param_pos, aaT_inv, ggT_inv
                )

            # restore original shapes
            if "weight" in param_pos:
                M_torch[param_pos["weight"]] = M_torch[param_pos["weight"]].view(
                    weight_shape
                )

        if return_tensor:
            M_torch = cat([rearrange(M, "k ... -> (...) k") for M in M_torch])

        return M_torch

    def torch_matvec(self, v_torch: ParameterMatrixType) -> ParameterMatrixType:
        """Apply the inverse of KFAC to a vector in PyTorch.

        This allows for matrix-vector products with the inverse KFAC approximation in
        PyTorch without converting tensors to numpy arrays, which avoids unnecessary
        device transfers when working with GPUs and flattening/concatenating.

        Args:
            v_torch: Vector for multiplication. If list of tensors, each entry has the
                same shape as a parameter, i.e. ``[p1.shape, p2.shape, ...]``.
                If tensor, has shape ``[D]``.

        Returns:
            Matrix-multiplication result ``KFAC⁻¹ @ v``. Return type is the same as the
            type of the input. If list of tensors, each entry has the same shape as a
            parameter, i.e. ``[p1.shape, p2.shape, ...]``. If tensor, has shape ``[D]``.

        Raises:
            ValueError: If the input tensor has the wrong data type.
        """
        if isinstance(v_torch, list):
            v_torch = [v_torch_i.unsqueeze(0) for v_torch_i in v_torch]
            result = self.torch_matmat(v_torch)
            return [res.squeeze(0) for res in result]
        elif isinstance(v_torch, Tensor):
            return self.torch_matmat(v_torch.unsqueeze(-1)).squeeze(-1)
        else:
            raise ValueError(
                f"Invalid input type: {type(v_torch)}. Expected list of tensors or tensor."
            )

    def _matmat(self, M: ndarray) -> ndarray:
        """Apply the inverse of KFAC to a matrix (multiple vectors).

        Args:
            M: Matrix for multiplication. Has shape ``[D, K]`` with some ``K``.

        Returns:
            Matrix-multiplication result ``KFAC⁻¹ @ M``. Has shape ``[D, K]``.
        """
        M_torch = self._A._preprocess(M)
        M_torch = self.torch_matmat(M_torch)
        return self._A._postprocess(M_torch)

    def state_dict(self) -> Dict[str, Any]:
        """Return the state of the inverse KFAC linear operator.

        Returns:
            State dictionary.
        """
        return {
            "A": self._A.state_dict(),
            # Attributes
            "damping": self._damping,
            "use_heuristic_damping": self._use_heuristic_damping,
            "min_damping": self._min_damping,
            "use_exact_damping": self._use_exact_damping,
            "cache": self._cache,
            "retry_double_precision": self._retry_double_precision,
            # Inverse Kronecker factors (if computed and cached)
            "inverse_input_covariances": self._inverse_input_covariances,
            "inverse_gradient_covariances": self._inverse_gradient_covariances,
        }

    def load_state_dict(self, state_dict: Dict[str, Any]):
        """Load the state of the inverse KFAC linear operator.

        Args:
            state_dict: State dictionary.
        """
        self._A.load_state_dict(state_dict["A"])

        # Set attributes
        self._damping = state_dict["damping"]
        self._use_heuristic_damping = state_dict["use_heuristic_damping"]
        self._min_damping = state_dict["min_damping"]
        self._use_exact_damping = state_dict["use_exact_damping"]
        self._cache = state_dict["cache"]
        self._retry_double_precision = state_dict["retry_double_precision"]

        # Set inverse Kronecker factors (if computed and cached)
        self._inverse_input_covariances = state_dict["inverse_input_covariances"]
        self._inverse_gradient_covariances = state_dict["inverse_gradient_covariances"]

    @classmethod
    def from_state_dict(
        cls, state_dict: Dict[str, Any], A: KFACLinearOperator
    ) -> "KFACInverseLinearOperator":
        """Load an inverse KFAC linear operator from a state dictionary.

        Args:
            state_dict: State dictionary.
            A: ``KFACLinearOperator`` whose inverse is formed.

        Returns:
            Linear operator of inverse KFAC approximation.
        """
        inv_kfac = cls(
            A,
            damping=state_dict["damping"],
            use_heuristic_damping=state_dict["use_heuristic_damping"],
            min_damping=state_dict["min_damping"],
            use_exact_damping=state_dict["use_exact_damping"],
            cache=state_dict["cache"],
            retry_double_precision=state_dict["retry_double_precision"],
        )
        inv_kfac.load_state_dict(state_dict)
        return inv_kfac



r"""Eigenvalues
===============

This example demonstrates how to compute a subset of eigenvalues of a linear
operator, using :func:`scipy.sparse.linalg.eigsh`. Concretely, we will compute
leading eigenvalues of the Hessian.

As always, imports go first.
"""

from contextlib import redirect_stderr
from io import StringIO
from typing import List, Tuple

import numpy
import scipy
import torch
from torch import nn

from curvlinops import HessianLinearOperator
from curvlinops.examples.functorch import functorch_hessian
from curvlinops.examples.utils import report_nonclose

# make deterministic
torch.manual_seed(0)
numpy.random.seed(0)

# %%
#
# Setup
# -----
#
# We will use synthetic data, consisting of two mini-batches, a small MLP, and
# mean-squared error as loss function.

N = 20
D_in = 7
D_hidden = 5
D_out = 3

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")

X1, y1 = torch.rand(N, D_in).to(DEVICE), torch.rand(N, D_out).to(DEVICE)
X2, y2 = torch.rand(N, D_in).to(DEVICE), torch.rand(N, D_out).to(DEVICE)

model = nn.Sequential(
    nn.Linear(D_in, D_hidden),
    nn.ReLU(),
    nn.Linear(D_hidden, D_hidden),
    nn.Sigmoid(),
    nn.Linear(D_hidden, D_out),
).to(DEVICE)
params = [p for p in model.parameters() if p.requires_grad]

loss_function = nn.MSELoss(reduction="mean").to(DEVICE)


# %%
#
# Linear operator
# ------------------
#
# We are ready to setup the linear operator. In this example, we will use the Hessian.

data = [(X1, y1), (X2, y2)]
H = HessianLinearOperator(model, loss_function, params, data)

# %%
#
# Leading eigenvalues
# -------------------
#
# Through :func:`scipy.sparse.linalg.eigsh`, we can obtain the leading
# :math:`k=3` eigenvalues.

k = 3
which = "LM"  # largest magnitude
top_k_evals, _ = scipy.sparse.linalg.eigsh(H, k=k, which=which)

print(f"Leading {k} Hessian eigenvalues: {top_k_evals}")

# %%
#
# Verifying results
# -----------------
#
# To double-check this result, let's compute the Hessian with
# :code:`functorch`, compute all its eigenvalues with
# :func:`scipy.linalg.eigh`, then extract the top :math:`k`.

H_functorch = (
    functorch_hessian(model, loss_function, params, data).detach().cpu().numpy()
)
evals_functorch, _ = scipy.linalg.eigh(H_functorch)
top_k_evals_functorch = evals_functorch[-k:]

print(f"Leading {k} Hessian eigenvalues (functorch): {top_k_evals_functorch}")

# %%
#
#  Both results should match.

print(f"Comparing leading {k} Hessian eigenvalues (linear operator vs. functorch).")
report_nonclose(top_k_evals, top_k_evals_functorch)

# %%
#
# :func:`scipy.sparse.linalg.eigsh` can also compute other subsets of
# eigenvalues, and also their associated eigenvectors. Check out its
# documentation for more!


# %%
#
# Power iteration versus ``eigsh``
# --------------------------------
#
# Here, we compare the query efficiency of :func:`scipy.sparse.linalg.eigsh` with the
# `power iteration <https://en.wikipedia.org/wiki/Power_iteration>`_ method, a simple
# method to compute the leading eigenvalues (in terms of magnitude). We re-use the im-
# plementation from the `PyHessian library <https://github.com/amirgholami/PyHessian>`_
# and adapt it to work with SciPy arrays rather than PyTorch tensors:


def power_method(
    A: scipy.sparse.linalg.LinearOperator,
    max_iterations: int = 100,
    tol: float = 1e-3,
    k: int = 1,
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """Compute the top-k eigenpairs of a linear operator using power iteration.

    Code modified from PyHessian, see
    https://github.com/amirgholami/PyHessian/blob/72e5f0a0d06142387fccdab2226b4c6bae088202/pyhessian/hessian.py#L111-L156

    Args:
        A: Linear operator of dimension ``D`` whose top eigenpairs will be computed.
        max_iterations: Maximum number of iterations. Defaults to ``100``.
        tol: Relative tolerance between two consecutive iterations that has to be
            reached for convergence. Defaults to ``1e-3``.
        k: Number of eigenpairs to compute. Defaults to ``1``.

    Returns:
        The eigenvalues as array of shape ``[k]`` in descending order, and their
        corresponding eigenvectors as array of shape ``[D, k]``.
    """
    eigenvalues = []
    eigenvectors = []

    def normalize(v: numpy.ndarray) -> numpy.ndarray:
        return v / numpy.linalg.norm(v)

    def orthonormalize(v: numpy.ndarray, basis: List[numpy.ndarray]) -> numpy.ndarray:
        for basis_vector in basis:
            v -= numpy.dot(v, basis_vector) * basis_vector
        return normalize(v)

    computed_dim = 0
    while computed_dim < k:
        eigenvalue = None
        v = normalize(numpy.random.randn(A.shape[0]))

        for _ in range(max_iterations):
            v = orthonormalize(v, eigenvectors)
            Av = A @ v

            tmp_eigenvalue = v.dot(Av)
            v = normalize(Av)

            if eigenvalue is None:
                eigenvalue = tmp_eigenvalue
            else:
                if abs(eigenvalue - tmp_eigenvalue) / (abs(eigenvalue) + 1e-6) < tol:
                    break
                else:
                    eigenvalue = tmp_eigenvalue

        eigenvalues.append(eigenvalue)
        eigenvectors.append(v)
        computed_dim += 1

    # sort in ascending order and convert into arrays
    eigenvalues = numpy.array(eigenvalues[::-1])
    eigenvectors = numpy.array(eigenvectors[::-1])

    return eigenvalues, eigenvectors


# %%
#
# Let's compute the top-3 eigenvalues via power iteration and verify they roughly match.
# Note that we are using a smaller :code:`tol` value than the PyHessian default value
# here to get better convergence, and we have to use relatively large tolerances for the
# comparison (which we didn't do when comparing :code:`eigsh` with :code:`eigh`).

top_k_evals_power, _ = power_method(H, tol=1e-4, k=k)
print(f"Comparing leading {k} Hessian eigenvalues (eigsh vs. power).")
report_nonclose(top_k_evals_functorch, top_k_evals_power, rtol=2e-2, atol=1e-6)

# %%
#
# This indicates that the power method achieves poorer accuracy than :code:`eigsh`. But
# does it therefore require fewer matrix-vector products? To answer this, let's turn on
# the linear operator's progress bar, which allows us to count the number of
# matrix-vector products invoked by both eigen-solvers:

H = HessianLinearOperator(model, loss_function, params, data, progressbar=True)

# determine number of matrix-vector products used by `eigsh`
with StringIO() as buf, redirect_stderr(buf):
    top_k_evals, _ = scipy.sparse.linalg.eigsh(H, k=k, which=which)
    # The tqdm progressbar will print "matmat" for each batch in a matrix-vector
    # product. Therefore, we need to divide by the number of batches
    queries_eigsh = buf.getvalue().count("matmat") // len(data)
print(f"eigsh used {queries_eigsh} matrix-vector products.")

# determine number of matrix-vector products used by power iteration
with StringIO() as buf, redirect_stderr(buf):
    top_k_evals_power, _ = power_method(H, k=k)
    # The tqdm progressbar will print "matmat" for each batch in a matrix-vector
    # product. Therefore, we need to divide by the number of batches
    queries_power = buf.getvalue().count("matmat") // len(data)
print(f"Power iteration used {queries_power} matrix-vector products.")

assert queries_power > queries_eigsh

# %%
#
# Sadly, the power iteration also does not offer computational benefits, consuming
# more matrix-vector products than :code:`eigsh`. While it is elegant and simple,
# it cannot compete with :code:`eigsh`, at least in the comparison provided here
# (note that we used a relative small tolerance for the power iteration, and it will
# likely deteriorate further if we decrease the tolerance).
#
# Therefore, we recommend using :code:`eigsh` for computing eigenvalues. This method
# becomes accessible because :code:`curvlinops` interfaces with SciPy's linear
# operators.



"""Contains LinearOperator implementation of the GGN."""

from __future__ import annotations

from collections.abc import MutableMapping
from typing import List, Tuple, Union

from backpack.hessianfree.ggnvp import ggn_vector_product_from_plist
from torch import Tensor, zeros_like

from curvlinops._base import _LinearOperator


class GGNLinearOperator(_LinearOperator):
    r"""GGN as SciPy linear operator.

    Consider the empirical risk

    .. math::
        \mathcal{L}(\mathbf{\theta})
        =
        c \sum_{n=1}^{N}
        \ell(f_{\mathbf{\theta}}(\mathbf{x}_n), \mathbf{y}_n)

    with :math:`c = \frac{1}{N}` for ``reduction='mean'`` and :math:`c=1` for
    ``reduction='sum'``. The GGN matrix is

    .. math::
        c \sum_{n=1}^{N}
        \left(
            \mathbf{J}_{\mathbf{\theta}}
            f_{\mathbf{\theta}}(\mathbf{x}_n)
        \right)^\top
        \left(
            \nabla_{f_\mathbf{\theta}(\mathbf{x}_n)}^2
            \ell(f_{\mathbf{\theta}}(\mathbf{x}_n), \mathbf{y}_n)
        \right)
        \left(
            \mathbf{J}_{\mathbf{\theta}}
            f_{\mathbf{\theta}}(\mathbf{x}_n)
        \right)\,.
    """

    def _matmat_batch(
        self, X: Union[Tensor, MutableMapping], y: Tensor, M_list: List[Tensor]
    ) -> Tuple[Tensor, ...]:
        """Apply the mini-batch GGN to a matrix.

        Args:
            X: Input to the DNN.
            y: Ground truth.
            M_list: Matrix to be multiplied with in list format.
                Tensors have same shape as trainable model parameters, and an
                additional leading axis for the matrix columns.

        Returns:
            Result of GGN multiplication in list format. Has the same shape as
            ``M_list``, i.e. each tensor in the list has the shape of a parameter and a
            leading dimension of matrix columns.
        """
        output = self._model_func(X)
        loss = self._loss_func(output, y)

        # collect matrix-matrix products per parameter
        result_list = [zeros_like(M) for M in M_list]

        num_vecs = M_list[0].shape[0]
        for n in range(num_vecs):
            col_n_list = ggn_vector_product_from_plist(
                loss, output, self._params, [M[n] for M in M_list]
            )
            for result, col_n in zip(result_list, col_n_list):
                result[n].add_(col_n)

        return tuple(result_list)

    def _adjoint(self) -> GGNLinearOperator:
        """Return the linear operator representing the adjoint.

        The GGN is real symmetric, and hence self-adjoint.

        Returns:
            Self.
        """
        return self



"""Utility linear operators."""

from __future__ import annotations

from numpy import einsum, einsum_path, ndarray, ones
from scipy.sparse.linalg import LinearOperator


class OuterProductLinearOperator(LinearOperator):
    """Linear operator for low-rank matrices of the form ``∑ᵢ cᵢ aᵢ aᵢᵀ``.

    ``cᵢ`` is the coefficient for the vector ``aᵢ``.
    """

    def __init__(self, c: ndarray, A: ndarray):
        """Store coefficients and vectors for low-rank representation.

        Args:
            c: Coefficients ``cᵢ``. Has shape ``[K]`` where ``K`` is the rank.
            A: Matrix of shape ``[D, K]``, where ``D`` is the linear operators
                dimension, that stores the low-rank vectors columnwise, i.e. ``aᵢ``
                is stored in ``A[:,i]``.
        """
        super().__init__(A.dtype, (A.shape[0], A.shape[0]))
        self._A = A
        self._c = c

        # optimize einsum
        self._equation = "ij,j,kj,k->i"
        self._operands = (self._A, self._c, self._A)
        placeholder = ones(self.shape[0])
        self._path = einsum_path(
            self._equation, *self._operands, placeholder, optimize="optimal"
        )[0]

    def _matvec(self, x: ndarray) -> ndarray:
        """Apply the linear operator to a vector.

        Args:
            x: Vector.

        Returns:
            Result of linear operator applied to the vector.
        """
        return einsum(self._equation, *self._operands, x, optimize=self._path)

    def _adjoint(self) -> OuterProductLinearOperator:
        """Return the linear operator representing the adjoint.

        An outer product is self-adjoint.

        Returns:
            Self.
        """
        return self


class Projector(OuterProductLinearOperator):
    """Linear operator for the projector onto the orthonormal basis ``{ aᵢ }``."""

    def __init__(self, A: ndarray):
        """Store orthonormal basis.

        Args:
            A: Matrix of shape ``[D, K]``, where ``D`` is the linear operators
                dimension, that stores the K orthonormal basis vectors columnwise,
                i.e. ``aᵢ`` is stored in ``A[:,i]``.
        """
        super().__init__(ones(A.shape[1]), A)
